# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Errors for the AP controller module."""


class BaseError(Exception):
  """Base error class for this controller module."""


class ConfigError(BaseError):
  """Configuration error."""


class HostapdStartError(BaseError):
  """Raised if it failed to start hostapd on the AP device."""


class SnifferManagerError(BaseError):
  """Raised if some failure related to sniffer manager happened."""


class SystemLogServiceError(BaseError):
  """Root error type for system log service."""

  def __init__(self, device, msg):
    new_msg = f'{repr(device)}::Service<SystemLog> {msg}'
    super().__init__(new_msg)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The manager managing packet capture instances on AP devices."""

import os
import signal
from typing import Any

from mobly import logger as mobly_logger
from mobly import runtime_test_info
from mobly import utils

from mobly.controllers.wifi.lib import constants
from mobly.controllers.wifi.lib import errors
from mobly.controllers.wifi.lib import iw_utils
from mobly.controllers.wifi.lib import sniffer_manager_base
from mobly.controllers.wifi.lib import wifi_configs

# Avoid directly importing cros_device, which causes circular dependencies.
OpenWrtDevice = Any

_FILE_TAG = 'sniffer'
_INTERFACE = 'monitor0'
_REMOTE_DIR_NAME = 'sniffer'

# Enable rotation with file size 50MB and at most 2 files for packet capturing.
# The packet capture process alternatively saves to 2 files. Once it reaches the
# file size limit it will start overwrite the other file.
# Note that the available RAM on Ubiquiti 6 lite AP is about 120MB, so
# `_PCAP_FILE_SIZE_MB` * `_KEEP_PCAP_FILE_NUM` must be smaller than 120MB.
_PCAP_FILE_SIZE_MB = 50
_KEEP_PCAP_FILE_NUM = 2
_KEEP_LATEST_PCAP_ARG = f'-W {_KEEP_PCAP_FILE_NUM} -C {_PCAP_FILE_SIZE_MB}'

# `tcpdump` filter to ignore QoS data frames.
_FILTER_IGNORE_QOS_DATA_FRAMES = r'not \(type data subtype qos-data\)'


class SnifferManager:
  """The class for managing sniffer instances on the AP device."""

  def __init__(self, device: 'OpenWrtDevice') -> None:
    self._device = device
    self._log = mobly_logger.PrefixLoggerAdapter(
        self._device.log,
        {
            mobly_logger.PrefixLoggerAdapter.EXTRA_KEY_LOG_PREFIX: (
                '[SnifferManager]'
            )
        },
    )

    self._remote_process = None
    self._capture_file_local_path = None
    self._local_work_dir = self._device.log_path
    self._timestamp = mobly_logger.get_log_file_timestamp()

  @property
  def is_alive(self) -> bool:
    """True if the service is alive; False otherwise."""
    return self._remote_process is not None

  @property
  def _remote_work_dir(self) -> Any:
    """The remote work directory of this object.

    We cannot initialize this in constructor because
    `self._device.remote_work_dir` can only be used after `device.initialize` is
    used.
    """
    remote_dir = os.path.join(self._device.remote_work_dir, _REMOTE_DIR_NAME)
    self._device.ssh.make_dirs(remote_dir)
    return remote_dir

  def start_capture(
      self,
      freq_config: wifi_configs.FreqConfig,
      capture_config: wifi_configs.PcapConfig | None = None,
  ) -> None:
    """Starts pacaket capture on the given channel."""
    capture_config = capture_config or wifi_configs.PcapConfig()
    self._log.debug(
        'Starting packet capture on frequency conf: %s, capture conf: %s',
        freq_config,
        capture_config,
    )
    self._assert_not_running()
    self._start_remote_process(freq_config, capture_config)
    self._print_interface_state()
    self._log.debug('Started packet capturing.')

  def _assert_not_running(self):
    """Asserts this manager is not running.

    Raises:
      errors.SnifferManagerError: if this manager is running.
    """
    if self.is_alive:
      raise errors.SnifferManagerError(
          'Running multiple sniffer instances is not allowed.'
      )

  def _start_remote_process(
      self,
      freq_config: wifi_configs.FreqConfig,
      capture_config: wifi_configs.PcapConfig,
  ) -> None:
    """Starts the remote process to capture packets on the device."""
    interface = _INTERFACE
    channel = freq_config.channel

    phys = iw_utils.get_all_phys(device=self._device)
    phy = iw_utils.get_phy_by_channel(phys, channel).name

    # Kill any existing tcpdump instances.
    self._device.ssh.execute_command(
        command=constants.Commands.KILLALL.format(name='tcpdump'),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
        ignore_error=True,
    )
    self._device.ssh.execute_command(
        command=constants.Commands.IW_DEV_DEL.format(interface=interface),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
        ignore_error=True,
    )
    self._device.ssh.execute_command(
        command=constants.Commands.IW_DEV_ADD_MONITOR.format(
            phy=phy, interface=interface
        ),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )
    self._device.ssh.execute_command(
        command=constants.Commands.IP_LINK_UP.format(interface=interface),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )
    cmd = sniffer_manager_base.gen_iw_set_freq_cmd(interface, freq_config)
    self._device.ssh.execute_command(
        command=cmd,
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )

    self._remove_capture_dir()

    self._timestamp = mobly_logger.get_log_file_timestamp()
    capture_file_remote_path = os.path.join(
        self._remote_work_dir, f'sniffer,{self._timestamp}.pcap'
    )

    capture_args = []
    if capture_config.keep_latest_packets:
      capture_args.append(_KEEP_LATEST_PCAP_ARG)
    if capture_config.ignore_qos_data_frames:
      capture_args.append(_FILTER_IGNORE_QOS_DATA_FRAMES)
    self._remote_process = self._device.ssh.start_remote_process(
        command=constants.Commands.START_TCPDUMP.format(
            interface=interface,
            file_path=capture_file_remote_path,
            args=' '.join(capture_args),
        ),
        get_pty=True,
    )

  def _remove_capture_dir(self):
    self._device.ssh.rm_dir(self._remote_work_dir)

  def _print_interface_state(self):
    """Prints interface state."""
    self._log.debug('Printing sniffer interface state.')
    self._device.ssh.execute_command(
        command=constants.Commands.IW_DEV_INFO.format(interface=_INTERFACE),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )
    self._device.ssh.execute_command(
        command=constants.Commands.IP_LINK_SHOW.format(interface=_INTERFACE),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )

  def __del__(self):
    self.teardown()

  def stop_capture(
      self, current_test_info: runtime_test_info.RuntimeTestInfo | None = None
  ):
    """Stops pacaket capture."""
    if not self.is_alive:
      self._log.warning(
          'Skip stopping this sniffer manager because it is not running.'
      )
      return
    self._log.debug('Stopping packet capturing.')
    self._stop_remote_process()
    if current_test_info is not None:
      local_dir = current_test_info.output_path
      self._pull_capture_files(local_dir)
    self._remove_capture_dir()
    self._log.debug('Stopped packet capturing.')

  def _stop_remote_process(self):
    """Stops the remote process."""
    if (proc := self._remote_process) is None:
      return

    self._remote_process = None
    proc.send_signal(signal_id=signal.SIGINT, assert_process_exit=True)
    self._log.debug('Sniffer process output: %s', proc.communicate())

  def _pull_capture_files(self, local_dir: str) -> None:
    """Pulls the capture files from the device to the host.

    If multiple capture files are found, this will use `mergecap` on host to
    merge them into one file. If `mergecap` is not installed, this will directly
    upload all capture files.

    Args:
      local_dir: the local directory to pull the capture files to.
    """
    # Print all capture files info for debugging.
    self._device.ssh.execute_command(
        command=f'ls -alh {self._remote_work_dir}',
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
        ignore_error=True,
    )
    pcap_files = list(self._device.ssh.list_dir(self._remote_work_dir))
    if not pcap_files:
      self._log.warning(
          'Packet capture file does not exist on device. This might be because'
          ' no packets were captured if the capturing time was too short.'
      )
      return

    self._device.ssh.pull_remote_directory(self._remote_work_dir, local_dir)
    pcap_local_paths = [os.path.join(local_dir, name) for name in pcap_files]

    # Skips merge if `mergecap` is not installed.
    which_mergecap = utils.run_command(['which', 'mergecap'])
    if which_mergecap[0] != 0:
      self._device.log.error(
          'Please install mergecap on your host machine. Otherwise, we will'
          ' directly upload all capture files without merging them into one'
          ' capture file.'
      )
      self._capture_file_local_path = pcap_local_paths[0]
      return

    new_file_path = self._merge_capture_files(pcap_local_paths, local_dir)
    self._capture_file_local_path = new_file_path

  def _merge_capture_files(self, pcap_local_paths: list[str], local_dir: str):
    """Merges multiple captures files into one capture file."""
    new_file_name = f'sniffer,merged,{self._timestamp}.pcap'
    new_file_path = os.path.join(local_dir, new_file_name)
    self._device.log.debug(
        'Merging following capture files into one file %s: %s',
        os.path.basename(new_file_path),
        ','.join(pcap_local_paths),
    )
    result = utils.run_command(
        ['mergecap', '-w', new_file_path] + pcap_local_paths
    )
    if result[0] != 0:
      raise errors.SnifferManagerError(
          f'Failed to merge capture files with result: {result}'
      )
    for pcap_local_path in pcap_local_paths:
      os.remove(pcap_local_path)
    return new_file_path

  def get_capture_file(self) -> str | None:
    """Gets the full path of the last capture."""
    return self._capture_file_local_path

  def teardown(self):
    """Tears down this manager object."""
    if not self.is_alive:
      return
    self.stop_capture()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mobly controller module for AP devices running on the OpenWrt system."""

from collections.abc import Iterator, Mapping, Sequence
import contextlib
import datetime
import itertools
import logging
import os
import time
from typing import Any

from mobly import logger as mobly_logger
from mobly import runtime_test_info
from mobly import utils
from mobly.controllers.android_device_lib import service_manager
import paramiko

from mobly.controllers.wifi.lib import ssh as ssh_lib
from mobly.controllers.wifi.lib import constants
from mobly.controllers.wifi.lib import device_info_utils
from mobly.controllers.wifi.lib import iw_utils
from mobly.controllers.wifi.lib import sniffer_manager
from mobly.controllers.wifi.lib import utils as wifi_utils
from mobly.controllers.wifi.lib import wifi_configs
from mobly.controllers.wifi.lib import wifi_manager
from mobly.controllers.wifi.lib.services import system_log_service


MOBLY_CONTROLLER_CONFIG_NAME = 'OpenWrtDevice'

_SSH_KEY_IDENTITY = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'data/testing_rsa'
)
_SSH_PORT = 22
_DEVICE_TAG = MOBLY_CONTROLLER_CONFIG_NAME

_DEVICE_REBOOT_WAIT = datetime.timedelta(seconds=10)
_BOOT_STATUS_CHECK_INTERVAL = datetime.timedelta(seconds=5)
_BOOT_STATUS_CHECK_TIMEOUT = datetime.timedelta(minutes=5)
_SSH_CONNECTION_TIMEOUT = datetime.timedelta(minutes=5)


_ERR_USE_AS_BOTH_AP_AND_SNIFFER = (
    "{device} It's not supported to use one OpenWrt device as both AP and"
    ' sniffer at the same time.'
)

_ERR_START_PACKET_CAPTURE_ARG_ERROR = (
    '{device} Exactly one of wifi_config and freq_config must be provided.'
)


class Error(Exception):
  """Error class for the OpenWrtDevice controller."""


def create(configs: list[dict[str, Any]]) -> list['OpenWrtDevice']:
  """Creates OpenWrt device instances."""
  if not configs:
    raise Error(f'Missing configuration {configs!r}.')
  devices = [OpenWrtDevice(config) for config in configs]
  devices = _initialize_devices(devices)
  return devices


def destroy(devices: list['OpenWrtDevice']) -> None:
  """Closes all created OpenWrt device instances."""
  for device in devices:
    try:
      device.teardown()
    except Exception:  # pylint: disable=broad-except
      logging.exception('Failed to clean up properly.')


def get_info(devices: Sequence['OpenWrtDevice']) -> Sequence[Mapping[str, Any]]:
  """Gets info from the OpenWrt device instances used in a test run.

  Args:
    devices: A list of OpenWrt device instances.

  Returns:
    A list of dict, each representing info for a device object.
  """
  return [d.device_info for d in devices]


def _initialize_devices(
    devices: Sequence['OpenWrtDevice'],
) -> list['OpenWrtDevice']:
  """Registers basic long running services on multiple OpenWrtDevice objects.

  Args:
    devices: A list of OpenWrtDevice objects.

  Returns:
    A list of initialized device objects.
  """
  initialized_devices = []
  for device in devices:
    try:
      device.initialize()
      initialized_devices.append(device)
    except Exception:  # pylint: disable=broad-except
      device.log.exception(
          'Failed to initialize AP device %s, ignoring it.', device
      )
  return initialized_devices


class OpenWrtDevice:
  """Mobly controller for AP devices running on the OpenWrt system.

  Attributes:
    ssh: The underlying SSH client object.
    serial: A string that identifies the ChromeOS device.
    log_path: A string that is the path where all logs collected on this device
      should be stored.
    debug_tag: A string that represents this ChromeOS device in the debug info.
    log: A logger adapted from root logger with an added prefix specific to a
      remote test machine. The prefix is "[OpenWrtDevice|<self.serial>] ".
    device_info: A collection of device information.
    services: The manager of long running services on the device.
    wifi_id_counter: The id counter of WiFi networks.
  """

  def __init__(self, config: dict[str, Any]):
    if 'hostname' not in config:
      raise Error(
          'Missing required field "hostname" in device configuration'
          f' {config!r}.'
      )
    self._hostname = config['hostname']
    self._username = config.get('username', constants.SSH_USERNAME)
    self._password = config.get('password', None)
    self._ssh_port = config.get('ssh_port', _SSH_PORT)
    self._skip_init_reboot = wifi_utils.convert_testbed_bool_value(
        config.get('skip_init_reboot', False)
    )
    self.serial = f'{self._hostname}:{self._ssh_port}'
    self._device_info = None

    log_path = getattr(logging, 'log_path', '/tmp/logs')
    log_filename = mobly_logger.sanitize_filename(
        f'{_DEVICE_TAG}_{self.serial}'
    )
    self.log_path = os.path.join(log_path, log_filename)
    utils.create_dir(self.log_path)

    self.log = mobly_logger.PrefixLoggerAdapter(
        logging.getLogger(),
        {
            mobly_logger.PrefixLoggerAdapter.EXTRA_KEY_LOG_PREFIX: (
                f'[{_DEVICE_TAG}|{self.serial}]'
            )
        },
    )

    self._remote_work_dir = None
    self._wifi_id_counter = itertools.count(0)

    self._ssh = self._create_ssh_client()
    self._wifi_manager = wifi_manager.WiFiManager(device=self)
    self.services = service_manager.ServiceManager(device=self)
    self._sniffer_manager = sniffer_manager.SnifferManager(device=self)

  def __repr__(self) -> str:
    return f'<{_DEVICE_TAG}|{self.serial}>'

  def _create_ssh_client(self) -> ssh_lib.SSHProxy:
    if self._password is None:
      return ssh_lib.SSHProxy(
          hostname=self._hostname,
          ssh_port=self._ssh_port,
          username=self._username,
          keyfile=_SSH_KEY_IDENTITY,
      )
    else:
      return ssh_lib.SSHProxy(
          hostname=self._hostname,
          ssh_port=self._ssh_port,
          username=self._username,
          password=self._password,
      )

  @property
  def wifi_id_counter(self) -> Iterator[int]:
    """The id counter of WiFi networks.

    This is a controller level counter and will not be reset until this
    controller object is teared down.

    Returns:
      The id counter.
    """
    return self._wifi_id_counter

  @property
  def remote_work_dir(self) -> str:
    """The path of the working directory on the AP device."""
    if self._remote_work_dir is None:
      time_str = mobly_logger.get_log_file_timestamp()
      self._remote_work_dir = os.path.join(
          constants.REMOTE_WORK_DIR, f'test-{time_str}'
      )
      self.make_dirs(self._remote_work_dir)
    return self._remote_work_dir

  def initialize(self):
    """Initializes this controller object.

    This method performs some one-off setup steps on the AP device.
    """
    os.chmod(_SSH_KEY_IDENTITY, 0o600)
    self._ssh.connect(
        open_sftp=False, timeout=_SSH_CONNECTION_TIMEOUT.total_seconds()
    )

    packages = constants.REQUIRED_PACKAGES_RELEASED_IMAGE
    if self.device_info.get(
        'device_name', None
    ) == constants.ApModel.BPIR3 and wifi_utils.is_using_custom_image(
        device=self
    ):
      packages = constants.REQUIRED_PACKAGES_BPIR3_AND_CROS_BUILT_IMAGE
    elif wifi_utils.is_using_openwrt_snapshot_image(
        self.device_info['release']
    ):
      packages = constants.REQUIRED_PACKAGES_SNAPSHOT_IMAGE

    for pkg in packages:
      self._install_package(pkg)

    if self._skip_init_reboot:
      self.log.info('Skipped reboot when initializing this controller object.')
      self._ssh.open_sftp()
      self._wifi_manager.initialize()
    else:
      self.reboot()

    self._register_syslog_service()

  def _install_package(self, package: str):
    """Installs a package on the device through `opkg`."""
    result = self.ssh.execute_command(
        command=constants.Commands.OPKG_LIST.format(package=package),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )
    if package in result:
      self.log.debug('Package %s is already installed.', package)
      return
    self.ssh.execute_command(
        command=constants.Commands.OPKG_UPDATE,
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )
    self.ssh.execute_command(
        command=constants.Commands.OPKG_INSTALL.format(package=package),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )

  def reboot(self) -> None:
    """Reboots the device.

    Generally one should use this method to reboot the device instead of
    directly calling `ssh.execute_command('sudo reboot')`. Because this method
    gracefully handles the teardown and restoration of running services.

    This method is blocking and only returns when the reboot has completed
    and the services restored.
    """
    self.log.info('Rebooting AP device...')
    with self.handle_reboot():
      # Use execute_command_async here to avoid getting stuck in dangling
      # ssh connection during rebooting.
      self.ssh.execute_command_async(command=constants.Commands.REBOOT)
      time.sleep(_DEVICE_REBOOT_WAIT.total_seconds())

  @contextlib.contextmanager
  def handle_reboot(self) -> Iterator[None]:
    """Properly manages the service life cycle when the device needs to reboot.

    The device can temporarily lose SSH connection due to user-triggered reboot.
    Use this function to make sure all Mobly components are properly stopped and
    restored afterwards.

    For sample usage, see self.reboot().

    Yields:
      None
    """
    self._wifi_manager.teardown()
    try:
      yield
    finally:
      self._ssh.disconnect()
      self._wait_for_boot_completion()
      self._wifi_manager.initialize()

  def _wait_for_boot_completion(self) -> None:
    """Waits for a ssh connection can be reestablished.

    Raises:
      Error: Raised if booting process timed out.
    """
    if not wifi_utils.wait_for_predicate(
        predicate=self._is_reboot_ready,
        timeout=_BOOT_STATUS_CHECK_TIMEOUT,
        interval=_BOOT_STATUS_CHECK_INTERVAL,
    ):
      raise Error(
          f'{repr(self)} Booting process timed out after'
          f' {_BOOT_STATUS_CHECK_TIMEOUT.total_seconds()} seconds.'
      )

  def ssh_connect(self):
    """Connects to the device through ssh with sftp enabled."""
    self._ssh.connect(
        open_sftp=True,
        timeout=_SSH_CONNECTION_TIMEOUT.total_seconds(),
    )

  def _is_reboot_ready(self) -> bool:
    """Returns whether the device is ready after reboot."""
    try:
      self.ssh_connect()
      self._ssh.execute_command(
          constants.Commands.CHECK_DEVICE_REBOOT_READY,
          timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
          ignore_error=False,
      )
      return True
    except (
        paramiko.ssh_exception.NoValidConnectionsError,
        paramiko.ssh_exception.SSHException,
        ConnectionResetError,
        TimeoutError,
        ssh_lib.ExecuteCommandError,
    ):
      # ssh connect may fail during certain period of booting
      # process, which is normal. Ignoring these errors.
      self.log.exception('Ignoring the exception when rebooting.')
      self._ssh.disconnect()
      return False

  def _register_syslog_service(self):
    try:
      self.services.register(
          alias='syslog', service_class=system_log_service.SystemLogService
      )
    except Exception:  # pylint: disable=broad-except
      self.log.exception(
          'Failed to register system log service, no system logs will be'
          ' collected for this device.'
      )

  def __del__(self):
    self.teardown()

  @property
  def device_info(self) -> Mapping[str, str]:
    """Information to be pulled into controller info in the test summary."""
    if self._device_info is None:
      self._device_info = device_info_utils.get_device_info(device=self) | {
          'serial': self.serial
      }
    return self._device_info

  @property
  def ssh(self) -> ssh_lib.SSHProxy:
    """The ssh connection to the AP device."""
    return self._ssh

  def make_dirs(self, remote_dir: str) -> None:
    """Recursively makes directories on the remote machine.

    Args:
      remote_dir: string, the remote directory.

    Raises:
      RuntimeError: a component in remote_dir is not a directory.
      SSHNotConnectedError: If sftp is not connected.
    """
    self.ssh.make_dirs(remote_dir)

  def push_file(
      self,
      local_src_filename: str,
      remote_dest_filename: str,
      change_permission: bool = False,
  ) -> None:
    """Pushes local file to the remote machine.

    Args:
      local_src_filename: the local file.
      remote_dest_filename: the destination file location in the remote machine.
      change_permission: whether to change the permission to 777 on remote
        destination.
    """
    self.ssh.push(local_src_filename, remote_dest_filename, change_permission)

  def start_wifi(
      self, config: wifi_configs.WiFiConfig
  ) -> wifi_configs.WifiInfo:
    if self._sniffer_manager.is_alive:
      raise Error(_ERR_USE_AS_BOTH_AP_AND_SNIFFER.format(device=self))
    return self._wifi_manager.start_wifi(config)

  def stop_wifi(self, wifi_info: wifi_configs.WifiInfo) -> None:
    self._wifi_manager.stop_wifi(wifi_info)

  def stop_all_wifi(self) -> None:
    self._wifi_manager.stop_all_wifi()

  def get_all_known_stations(
      self, wifi_info: wifi_configs.WifiInfo
  ) -> Sequence[iw_utils.Station]:
    """Gets all the known stations associated with the given WiFi network.

    Clients connected to WiFi network are part of stations.

    Args:
      wifi_info: The WiFi network to query the known stations.

    Returns:
      The sequence of all known stations.
    """
    return iw_utils.get_all_known_stations(
        device=self, interface=wifi_info.interface
    )

  def get_station_info(
      self, wifi_info: wifi_configs.WifiInfo, mac_address: str
  ) -> iw_utils.Station:
    """Gets info for the station with the given MAC address.

    Clients connected to WiFi network are part of stations.

    Args:
      wifi_info: The WiFi network to query the known stations.
      mac_address: The MAC address of the station.

    Returns:
      The information for the specified station.

    Raises:
      iw_utils.NoSuchStationError: Failed to find the specified station.
    """
    return iw_utils.get_station_info(
        device=self, interface=wifi_info.interface, mac_address=mac_address
    )

  def get_all_wifi_ssid(self) -> Sequence[str]:
    """Gets all currently broadcasting Wi-Fi SSIDs."""
    return [
        interface.ssid
        for interface in iw_utils.get_all_interfaces(self)
        if interface.type == 'AP' and interface.ssid is not None
    ]

  def start_packet_capture(
      self,
      wifi_config: wifi_configs.WiFiConfig | None = None,
      freq_config: wifi_configs.FreqConfig | None = None,
      capture_config: wifi_configs.PcapConfig | None = None,
  ):
    """Starts pacaket capture on a specific channel.

    You need to provide either wifi_config or freq_config. If you provide
    `freq_config`, this will monitor the frequency band specified by it. If you
    provide `wifi_config`, this method will extract channel info from it and
    monitor the same channel that is used by the WiFi network.

    Args:
      wifi_config: The WiFi network to capture the packets.
      freq_config: The frequency to capture the packets.
      capture_config: The configuration to control the packet capture process.

    Raises:
      Error: If both wifi_config and freq_config are provided or not provided.
    """
    if sum([wifi_config is not None, freq_config is not None]) != 1:
      raise Error(_ERR_START_PACKET_CAPTURE_ARG_ERROR.format(device=self))
    if self._wifi_manager.is_alive:
      raise Error(_ERR_USE_AS_BOTH_AP_AND_SNIFFER.format(device=self))

    if wifi_config is not None:
      freq_config = wifi_configs.get_freq_config(wifi_config)
    self._sniffer_manager.start_capture(
        freq_config=freq_config, capture_config=capture_config
    )

  def stop_packet_capture(
      self, current_test_info: runtime_test_info.RuntimeTestInfo | None = None
  ):
    """Stops pacaket capture."""
    self._sniffer_manager.stop_capture(current_test_info=current_test_info)

  def get_capture_file(self) -> str | None:
    """Gets the full path of the last capture."""
    return self._sniffer_manager.get_capture_file()

  def teardown(self):
    """Tears the device object down."""
    self.log.info('Tearing down the controller.')
    self._sniffer_manager.teardown()
    self._wifi_manager.teardown()
    self._ssh.disconnect()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The module for managing the lifecycle of WiFi networks.

This is entry point for WiFi network level management, i.e., start / stop / get
WiFi network status.
"""

import contextlib
import dataclasses
import datetime
from typing import Any

from mobly import logger as mobly_logger

from mobly.controllers.wifi.lib import ssh as ssh_lib
from mobly.controllers.wifi.lib import constants
from mobly.controllers.wifi.lib import dhcp_manager
from mobly.controllers.wifi.lib import errors
from mobly.controllers.wifi.lib import hostapd_manager
from mobly.controllers.wifi.lib import iw_utils
from mobly.controllers.wifi.lib import utils
from mobly.controllers.wifi.lib import wifi_configs

# Avoid directly importing cros_device, which causes circular dependencies
OpenWrtDevice = Any


WIFI_START_WAIT_TIME = datetime.timedelta(seconds=30)
# DFS channels require 60s extra start time to check channel availability.
WIFI_START_WAIT_TIME_DFS = datetime.timedelta(seconds=180)


@dataclasses.dataclass
class _WiFiComponents:
  """The data class that combines all components related to one WiFi network.

  Attributes:
    info: The WiFi network information.
    config: The user specified WiFi configurations.
    hostapd_manager: The hostapd manager instance.
    dhcp_manager: The dhcp manager instance.
  """

  info: wifi_configs.WifiInfo
  config: wifi_configs.WiFiConfig
  hostapd_manager: hostapd_manager.HostapdManager
  dhcp_manager: dhcp_manager.DhcpManager | None


class WiFiManager:
  """The class for managing the lifecycle of WiFi networks.

  This class is the entry point of WiFi network level management, i.e., start /
  stop / get WiFi status. To manage WiFi networks, this class utilizes other
  modules like hostapd_manager, dhcp_manager to manage instances running on AP
  devices.
  """

  _running_wifis: dict[int, _WiFiComponents]

  def __init__(self, device: 'OpenWrtDevice'):
    self._device = device
    self._id_counter = device.wifi_id_counter
    self._running_wifis = {}
    self._wan_interface = constants.WAN_INTERFACE

    self._log = mobly_logger.PrefixLoggerAdapter(
        device.log,
        {
            mobly_logger.PrefixLoggerAdapter.EXTRA_KEY_LOG_PREFIX: (
                '[WiFiManager]'
            )
        },
    )

  def initialize(self):
    """Initializes the AP device to be ready for starting WiFi networks.

    This method should be called after each AP device reboot.
    """
    # Kill any existing hostapd instances.
    self._device.ssh.execute_command(
        command=constants.Commands.KILLALL.format(name=constants.HOSTAPD),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
        ignore_error=True,
    )

    # We don't need to remove these firewall rule modifications in teardown
    # phase because device reboot automatically reset them.
    # Enable the kernel to route packets between different networks.
    self._device.ssh.execute_command(
        command=constants.Commands.FIREWALL_ENABLE_IP_FORWARD,
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )

    if utils.is_using_openwrt_snapshot_image(
        self._device.device_info['release']
    ):
      return

    # TODO: Temp workaround for new firewall rule.
    if utils.is_new_firewall_rule_version(self._device.device_info['release']):
      return

    # Enable NAT, i.e., outbound traffic through the WAN interface will have its
    # source addr replaced with the WAN interface.
    self._device.ssh.execute_command(
        command=constants.Commands.FIREWALL_ENABLE_NAT.format(
            interface=self._wan_interface
        ),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )

  @property
  def is_alive(self) -> bool:
    """True if there are any running WiFi networks, False otherwise."""
    return bool(self._running_wifis)

  def start_wifi(
      self, config: wifi_configs.WiFiConfig
  ) -> wifi_configs.WifiInfo:
    """Starts a WiFi network with the given configurations."""
    wifi_id = next(self._id_counter)
    self._log.debug(
        'Starting a WiFi network (id=%d) with config: %s', wifi_id, config
    )

    try:
      return self._start_wifi(wifi_id, config)
    except (ssh_lib.SSHRemoteError, errors.BaseError):
      self._log.error(
          'Cleaning up allocated resources for WiFi network (id=%d) due to'
          ' start failure.',
          wifi_id,
      )
      if (component := self._running_wifis.pop(wifi_id, None)) is not None:
        with contextlib.suppress(errors.BaseError, ssh_lib.SSHRemoteError):
          self._stop_wifi(component=component)
      raise

  def _start_wifi(
      self, wifi_id: int, config: wifi_configs.WiFiConfig
  ) -> wifi_configs.WifiInfo:
    """Performs the device operations to start a specified WiFi network."""
    self._print_debug_info_before_starting_wifi()
    self._set_country_code(config.country_code)
    # Parse phy info after setting country code because hardware capabilities
    # are affected by country code.
    phys = iw_utils.get_all_phys(self._device)
    phy = iw_utils.get_phy_by_channel(phys, channel=config.channel)
    interface = self._create_virtual_interface(wifi_id, phy)

    hostapd_manager_obj = hostapd_manager.HostapdManager(
        device=self._device,
        wifi_id=wifi_id,
        phy=phy,
        interface=interface,
        wifi_config=config,
    )
    wifi_info = hostapd_manager_obj.start()
    self._running_wifis[wifi_id] = _WiFiComponents(
        info=wifi_info,
        config=config,
        hostapd_manager=hostapd_manager_obj,
        dhcp_manager=None,
    )

    if config.access_wan_through_nat:
      dhcp_manager_obj = dhcp_manager.DhcpManager(
          device=self._device, wifi_id=wifi_id, iface=interface
      )
      dhcp_manager_obj.start()
      self._running_wifis[wifi_id].dhcp_manager = dhcp_manager_obj

    self._modify_firewall_rules(
        interface,
        action=constants.IptablesAction.INSERT,
        access_wan_through_nat=config.access_wan_through_nat,
    )

    self._set_txpower(interface, config)

    self._log.debug(
        'Started WiFi network %d with config: %s', wifi_info.id, config
    )

    # Print device status info for debugging.
    self._device.ssh.execute_command(
        command=constants.Commands.IW_DEV_INFO.format(interface=interface),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )
    self._device.ssh.execute_command(
        command=constants.Commands.IW_REG_GET,
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )

    return wifi_info

  def _set_txpower(self, interface: str, config: wifi_configs.WiFiConfig):
    """Sets the transmit power of the wireless interface on the AP device."""
    if config.maximum_txpower_dbm is None:
      return
    txpower_mbm = config.maximum_txpower_dbm * 100
    self._device.ssh.execute_command(
        command=constants.Commands.IW_DEV_SET_MAXIMUM_TXPOWER.format(
            interface=interface,
            txpower_mbm=txpower_mbm,
        ),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )

  def _print_debug_info_before_starting_wifi(self):
    """Prints debug information before starting each WiFi network."""
    self._device.ssh.execute_command(
        command=constants.Commands.GET_PROCESS_BY_NAME.format(
            name=constants.HOSTAPD
        ),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
        ignore_error=True,
    )

  def _set_country_code(self, country_code: str):
    """Sets country code to AP devices."""
    self._device.ssh.execute_command(
        command=constants.Commands.IW_REG_SET.format(
            country_code=country_code,
        ),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )

  def _create_virtual_interface(self, wifi_id: int, phy: iw_utils.Phy) -> str:
    """Creates a new virtual interface using the given hardware device."""
    interface = f'managed{wifi_id}'
    self._device.ssh.execute_command(
        command=constants.Commands.IW_DEV_DEL.format(interface=interface),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
        ignore_error=True,
    )
    self._device.ssh.execute_command(
        command=constants.Commands.IW_DEV_ADD.format(
            phy=phy.name, interface=interface
        ),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )
    return interface

  def _modify_firewall_rules(
      self,
      interface: str,
      action: constants.IptablesAction,
      access_wan_through_nat: bool,
  ):
    """Modifies firewall rules to forward traffic between the WLAN and WAN."""
    if utils.is_using_openwrt_snapshot_image(
        self._device.device_info['release']
    ):
      return

    # TODO: Temp workaround for new firewall rule.
    if utils.is_new_firewall_rule_version(self._device.device_info['release']):
      if access_wan_through_nat and not utils.is_using_custom_image(
          self._device
      ):
        raise errors.ConfigError(
            'Config access_wan_through_nat=True is not supported under current'
            'official OpenWrt image and version'
            f' {self._device.device_info["release"]}'
        )
      return

    wan_interface = self._wan_interface
    wireless_interface = interface

    # The rule that allows all packets from wireless interface to WAN interface.
    self._device.ssh.execute_command(
        command=constants.Commands.FIREWALL_FORWARD_KNOWN_TRAFFIC.format(
            action=action,
            in_interface=wan_interface,
            out_interface=wireless_interface,
        ),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )
    # The rule that only allows packets that are from WAN interface to wireless
    # interface and are part of a connection that has already been established,
    # such as a TCP connection.
    self._device.ssh.execute_command(
        command=constants.Commands.FIREWALL_FORWARD_TRAFFIC.format(
            action=action,
            in_interface=wireless_interface,
            out_interface=wan_interface,
        ),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
    )

  def __del__(self):
    self.teardown()

  def teardown(self):
    """Tears this WiFi manager down and stops all running WiFi."""
    self.stop_all_wifi()

  def stop_all_wifi(self):
    """Stops all running WiFi."""
    self._log.debug('Stopping all running WiFi networks.')
    wifis = self._running_wifis
    self._running_wifis = {}
    for wifi_id, component in list(wifis.items()):
      try:
        self._stop_wifi(component=component)
      except (ssh_lib.RemoteTimeoutError, ssh_lib.SSHRemoteError):
        self._log.exception(
            'Ignoring the exception when trying to stop WiFi %s', wifi_id
        )

  def stop_wifi(self, wifi_info: wifi_configs.WifiInfo):
    """Stops the given WiFi network."""
    component = self._running_wifis.pop(wifi_info.id, None)
    if component is None:
      self._log.debug(
          'WiFi network with following WifiInfo does not exist or is already'
          ' stopped: %s',
          wifi_info,
      )
      return

    self._stop_wifi(component=component)

  def _stop_wifi(self, *, component: _WiFiComponents):
    """Stops all the given WiFi components related to one WiFi network."""
    self._log.debug('Stopping WiFi network with id %d', component.info.id)

    # Remove added firewall rules.
    self._modify_firewall_rules(
        component.info.interface,
        action=constants.IptablesAction.DELETE,
        access_wan_through_nat=component.config.access_wan_through_nat,
    )

    # Stop hostapd manager.
    try:
      component.hostapd_manager.stop()
    except ssh_lib.SSHRemoteError as e:
      # We need to stop other components when hostapd_manager stop failed.
      if component.dhcp_manager is not None:
        component.dhcp_manager.stop()
      raise e

    if component.dhcp_manager is not None:
      component.dhcp_manager.stop()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The service for streaming the system logs on AP device to host."""

import datetime
import pathlib
from typing import Any

from mobly import logger as mobly_logger
from mobly import runtime_test_info
from mobly import utils
from mobly.controllers.android_device_lib.services import base_service

from mobly.controllers.wifi.lib import file_clipper
from mobly.controllers.wifi.lib import constants
from mobly.controllers.wifi.lib import errors

# Avoid directly importing cros_device, which causes circular dependencies.
OpenWrtDevice = Any

_FILE_TAG = 'openwrt_syslog'
_REMOTE_PROC_STOP_WAIT_TIME = datetime.timedelta(seconds=30)


class SystemLogService(base_service.BaseService):
  """The service for streaming the system logs on AP device to host."""

  def __init__(self, device: 'OpenWrtDevice', configs: Any = None) -> None:
    del configs  # Unused.
    self._device = device
    self._configs = None
    self._log = mobly_logger.PrefixLoggerAdapter(
        self._device.log,
        {
            mobly_logger.PrefixLoggerAdapter.EXTRA_KEY_LOG_PREFIX: (
                '[SystemLogService]'
            )
        },
    )

    self._remote_process = None
    host_log_dir = pathlib.Path(self._device.log_path)
    self._host_log_path = pathlib.Path(host_log_dir, f'{_FILE_TAG}.log')
    self._excerpt_generator = file_clipper.FileClipper(self._host_log_path)

  @property
  def is_alive(self) -> bool:
    """True if the service is alive; False otherwise."""
    return self._remote_process is not None

  def start(self) -> None:
    """Starts streaming the specified logs to host."""
    self._log.debug('Starting.')
    self._assert_not_running()
    self._start_log_subprocess()
    self._log.debug('Started.')

  def _assert_not_running(self):
    """Asserts the logcat service is not running.

    Raises:
      errors.SystemLogServiceError: if the logcat service is running.
    """
    if self.is_alive:
      raise errors.SystemLogServiceError(
          self._device,
          'System log service is already running, cannot start again.',
      )

  def _start_log_subprocess(self) -> None:
    """Starts the subprocess which streams logs to host."""
    self._remote_process = self._device.ssh.start_remote_process(
        command=constants.Commands.LOGREAD,
        get_pty=True,
        output_file_path=str(self._host_log_path),
    )

  def __del__(self):
    self.stop()

  def stop(self) -> None:
    """Stops streaming device system logs to host.

    Raises:
      errors.SystemLogServiceError: raised if the syslog subprocess has been
        stopped before this service.
    """
    self._log.debug('Stopping.')
    self._excerpt_generator.close()
    self._stop_log_subprocess()
    self._log.debug('Stopped.')

  def _stop_log_subprocess(self):
    if (proc := self._remote_process) is None:
      return

    self._remote_process = None
    proc.terminate(
        timeout=_REMOTE_PROC_STOP_WAIT_TIME.total_seconds(),
        assert_process_exit=True,
    )

  def create_output_excerpts(
      self, test_info: runtime_test_info.RuntimeTestInfo
  ) -> list[Any]:
    """Creates excerpts for system logs and returns the excerpt paths."""
    self._log.debug('Creating output excerpts.')
    dest_path = test_info.output_path
    utils.create_dir(dest_path)
    timestamp = mobly_logger.get_log_file_timestamp()

    filename = f'{_FILE_TAG},{self._device.serial},{timestamp}.log'
    excerpts_path = pathlib.Path(dest_path, filename)
    self._excerpt_generator.clip_new_content(excerpts_path)

    return [excerpts_path]



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A wrapper of paramiko.SSHClient to interact with test machine remotely."""

from __future__ import annotations
from collections.abc import Iterable, Mapping
import contextlib
import dataclasses
import errno
import logging
import os
import pathlib
import signal
import socket
import stat
import threading
import time
from typing import Optional, cast

from mobly import logger as mobly_logger
import paramiko
from paramiko import channel
from paramiko import sftp_attr
from zmq.ssh import forward

from mobly.controllers.wifi.lib import channel_file_streamer

_SFTP_NOT_CONNECTED_ERROR_MESSAGE = (
    'Cannot use SFTP without opening the SFTP session. Probably because the'
    ' SSH connection is not established or SFTP session is not opened when'
    ' establishing the SSH connection.'
)
_SFTP_CANNOT_BE_OPENED_DUPLICATEDLY = (
    'Unable to open a new SFTP session while another is active.'
)

_SIGNAL_CMD_TIMEOUT_SEC = 4
_WAIT_PROC_EXIT_TIMEOUT_SEC = 30


@dataclasses.dataclass
class CommandResults:
  """A container to collect full command results."""
  exit_code: int = 0
  output: str = ''
  error: str = ''


class Error(Exception):
  """Base error type for ssh module."""

  def __init__(self, ssh: SSHProxy, message: str) -> None:
    super().__init__(f'{repr(ssh)} {message}')


class SSHRemoteError(Error):
  """Raised when a ssh operation encounters an error."""
  pass


class RemotePathDoesNotExistError(Error):
  """Raised when a path doesn't exist on the remote machine."""
  pass


class ExecuteCommandError(Error):
  """Raised when a ssh command encounters an error."""

  _COMMAND_EXCEPTION_TEMPLATE = """
  Call exited with non-zero return code of "{return_code:d}".
  ****************************Call***************************
  {command}
  ****************************Stdout*************************
  {stdout}
  ****************************Stderr*************************
  {stderr}
  **********************End of error message*****************
  """

  def __init__(self, ssh: SSHProxy, command: str,
               command_results: CommandResults) -> None:
    message = self._COMMAND_EXCEPTION_TEMPLATE.format(
        command=command,
        return_code=command_results.exit_code,
        stdout=command_results.output,
        stderr=command_results.error,
    )
    super().__init__(ssh, message)


class SSHNotConnectedError(Error):
  """Raised when the SSH client is not connected."""


class SFTPAlreadyConnectedError(Error):
  """Raised when trying to open a SFTP session while one is already opened."""


class RemoteTimeoutError(Error):
  """Raised when a remote process did not complete in its given time.."""
  pass


class PortForwardingError(Error):
  """Raised when a port forwarding operation encounters an error."""


class SSHProxy:
  """SSH client to interact with test machine.

  Attributes:
    log: A logger adapted from root logger with an added prefix specific to a
      remote test machine. The prefix is "[SSHProxy| hostname:ssh_port] ".
    ssh_client: the underlying Paramiko SSHClient object.
  """

  def __init__(
      self,
      hostname: str,
      ssh_port: int = 22,
      username: str = 'root',
      password: Optional[str] = None,
      keyfile: Optional[str] = None,
      allow_agent: bool = False,
      proxy_command: str | None = None,
  ):
    """Initializes the SSH client instance.

    Args:
      hostname: the IP address of the test machine.
      ssh_port: the ssh port of the test machine.
      username: the user name to log in to test machine.
      password: the password to log in to test machine.
      keyfile: a local path to a private key file.
      allow_agent: allow use of ssh-agent for underlying ssh_client object.
      proxy_command: the ProxyCommand string to use when connecting to the
      remote machine.
    """
    self._hostname = hostname
    self._username = username
    self._password = password
    self._ssh_port = ssh_port
    self._ssh_keyfile = keyfile
    self._allow_agent = allow_agent
    self._port_forward_servers: dict[int, forward.ForwardServer] = {}
    self._proxy_command = (
        paramiko.ProxyCommand(proxy_command)
        if proxy_command is not None
        else None
    )

    self._sftp = None
    self.log = mobly_logger.PrefixLoggerAdapter(
        logging.getLogger(),
        {
            mobly_logger.PrefixLoggerAdapter.EXTRA_KEY_LOG_PREFIX:
                f'[SSHProxy|{self._hostname}:{self._ssh_port}]'
        }
    )
    self.ssh_client = paramiko.SSHClient()
    self.ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    # Suppress Paramiko debug logging. In some circumstances the Paramiko client
    # will continue to log debug messages after it has been closed. This can
    # cause crashes on system shutdown in Python3.
    logging.getLogger('paramiko').setLevel(logging.WARNING)

  def __repr__(self):
    return f'<SSHProxy|{self._hostname}:{self._ssh_port}>'

  def connect(
      self,
      timeout: float | None = None,
      banner_timeout: float | None = None,
      open_sftp: bool = True,
  ) -> None:
    """Connects to the test machine.

    Arguments:
      timeout: Connection timeout in seconds.
      banner_timeout: Timeout for the SSH banner to be presented.
      open_sftp: Whether to open a SFTP session after connection established.
        Without SFTP session, methods that related to remote file operations
        cannot be used.
    """
    self.log.info('Connecting to %s:%d', self._hostname, self._ssh_port)

    self.ssh_client.connect(
        self._hostname,
        port=self._ssh_port,
        username=self._username,
        password=self._password,
        allow_agent=self._allow_agent,
        key_filename=self._ssh_keyfile,
        sock=self._proxy_command,
        timeout=timeout,
        banner_timeout=banner_timeout,
    )
    if open_sftp:
      self.open_sftp()

  def open_sftp(self) -> None:
    """Opens an SFTP session.

    This is required by file operations provided in this class.

    Raises:
      SFTPAlreadyConnectedError: Raised if trying to open a SFTP session while
        one is already opened.
    """
    if self._sftp is not None:
      raise SFTPAlreadyConnectedError(
          ssh=self, message=_SFTP_CANNOT_BE_OPENED_DUPLICATEDLY
      )
    self._sftp = self.ssh_client.open_sftp()

  def disconnect(self):
    """Disconnects from the test machine and cleans up."""
    self.log.info('Disconnecting from %s', self._hostname)

    for forwarded_local_port in list(self._port_forward_servers):
      self.stop_port_forwarding(forwarded_local_port)

    if self._sftp:
      self._sftp.close()
      self._sftp = None

    self.ssh_client.close()

  def forward_port(self, remote_port: int, local_port: int = 0) -> int:
    """Sets up an SSH tunnel to port forward between client and remote host.

    The tunnel will run in a separate thread.

    Args:
      remote_port: The port on the remote host to forward all data to.
      local_port: The port on the local machine from which to forward all data.
        If 0 is passed then an unused port is picked.

    Returns:
      The local port that is used to open the SSH tunnel.

    Raises:
      PortForwardingError: Raised if trying to forward different remote ports
        to the same local port.
    """
    # Modifications were required in order to get a handle to the server object
    # which is used to stop the server during clean up.

    class SubHandler(forward.Handler):
      """Configures the local forward server."""
      chain_host = self._hostname
      chain_port = remote_port
      ssh_transport = self.ssh_client.get_transport()

    if local_port != 0 and local_port in self._port_forward_servers:
      forwarded_server = self._port_forward_servers[local_port]
      forwarded_remote_port = cast(
          SubHandler, forwarded_server.RequestHandlerClass
      ).chain_port

      if remote_port == forwarded_remote_port:
        return local_port

      error_string = (f'Local port {local_port} is in use for forwarding '
                      f'remote port {forwarded_remote_port}. Please choose '
                      'an unused local port, or simply set it to 0.')
      raise PortForwardingError(ssh=self, message=error_string)

    port_forward_server = (
        forward.ForwardServer(('127.0.0.1', local_port), SubHandler))

    # Start the server on a separate thread.
    thread = threading.Thread(
        target=port_forward_server.serve_forever, daemon=True)
    thread.start()

    _, forwarded_local_port = port_forward_server.server_address
    self._port_forward_servers[forwarded_local_port] = port_forward_server

    self.log.debug('Forwarded address %s:%d to local port %d', self._hostname,
                   remote_port, forwarded_local_port)
    return forwarded_local_port

  def stop_port_forwarding(self, local_port: int) -> None:
    """Stops port forwarded in function `forward_port`.

    Args:
      local_port: The forwarded port on the remote device.

    Raises:
      PortForwardingError: Raised if trying to stop forwarding to a local port
        which isn't being forwarding.
    """
    if local_port not in self._port_forward_servers:
      error_message = (f'Trying to stop forwarding to local port {local_port} '
                       "which isn't being forwarding.")
      raise PortForwardingError(ssh=self, message=error_message)

    server = self._port_forward_servers.pop(local_port)
    server.shutdown()
    server.server_close()

  def push_dir(self,
               local_src_dir: str,
               remote_dest_dir: str,
               change_permission: bool = False) -> None:
    """Pushes local directory recursively to the remote machine.

    Args:
      local_src_dir: the local directory to be copied to the remote machine.
      remote_dest_dir: the destination directory in the remote machine.
      change_permission: whether to change the permission to 777 on remote
        destination.
    """
    self.log.debug(
        'Pushing dir %s to test machine dir at %s',
        local_src_dir,
        remote_dest_dir,
    )
    self.make_dirs(remote_dest_dir)
    for (dir_path, _, file_names_list) in os.walk(local_src_dir):
      for file_name in file_names_list:
        local_file = os.path.join(dir_path, file_name)
        self.push(
            local_file,
            local_file.replace(local_src_dir, remote_dest_dir, 1),
            change_permission=change_permission)

  def move_dir(self, remote_current_location: str,
               remote_destination_location: str) -> None:
    """Renames a folder on the remote machine.

    Args:
      remote_current_location: The current name of the folder.
      remote_destination_location: The desired name of the folder. Exception
        will be thrown if remote_destination_location exists.

    Raises:
      IOError: if remote_destination_location is a folder.
      SSHNotConnectedError: If sftp is not connected.
    """
    if self._sftp is None:
      raise SSHNotConnectedError(
          ssh=self,
          message=_SFTP_NOT_CONNECTED_ERROR_MESSAGE,
      )
    self._sftp.rename(remote_current_location, remote_destination_location)

  def push(self,
           local_src_filename: str,
           remote_dest_filename: str,
           change_permission: bool = False) -> None:
    """Pushes local file to the remote machine.

    Args:
      local_src_filename: the local file.
      remote_dest_filename: the destination file location in the remote machine.
      change_permission: whether to change the permission to 777 on remote
        destination.

    Raises:
      FileNotFoundError: If the local file is missing.
      SSHNotConnectedError: If sftp is not connected.
    """
    if self._sftp is None:
      raise SSHNotConnectedError(
          ssh=self,
          message=_SFTP_NOT_CONNECTED_ERROR_MESSAGE,
      )

    self.make_dirs(os.path.dirname(remote_dest_filename))
    self._sftp.put(local_src_filename, remote_dest_filename)
    # Restore file permissions in linux and mac
    if change_permission:
      self._sftp.chmod(remote_dest_filename,
                       (os.stat(local_src_filename).st_mode & 0o777))

  def pull(self, remote_src_filepath: str, local_dest_filepath: str) -> None:
    """Pulls a remote file from the client as a local file.

    Args:
      remote_src_filepath: the remote source file name.
      local_dest_filepath: the local destination file name.

    Raises:
      RemotePathDoesNotExistError: If the remote file is missing.
      SSHNotConnectedError: If sftp is not connected.
    """
    if not self.exists(remote_src_filepath):
      raise RemotePathDoesNotExistError(
          self, f'The file to pull ({remote_src_filepath}) does not exist.')

    local_dir_path = os.path.dirname(local_dest_filepath)
    if not os.path.exists(local_dir_path):
      os.makedirs(local_dir_path)

    if self._sftp is None:
      raise SSHNotConnectedError(
          ssh=self,
          message=_SFTP_NOT_CONNECTED_ERROR_MESSAGE,
      )
    self._sftp.get(remote_src_filepath, local_dest_filepath)

  def pull_to_directory(self, remote_src_filepath: str,
                        local_directory: str) -> None:
    """Pulls a remote file from the client into a local directory.

    Args:
      remote_src_filepath: the remote source file path.
      local_directory: the path to the local destination directory.
    """
    filename = os.path.basename(remote_src_filepath)
    self.pull(remote_src_filepath, os.path.join(local_directory, filename))

  def _get_file_paths_in_remote_directory(
      self,
      remote_dir: pathlib.PurePosixPath) -> Iterable[pathlib.PurePosixPath]:
    """Yields the path to all files within a remote directory recursively.

    Args:
      remote_dir: The remote directory path.

    Yields:
      The path to all files in the directory.

    Raises:
      SSHNotConnectedError: If sftp is not connected.
    """
    if self._sftp is None:
      raise SSHNotConnectedError(
          ssh=self,
          message=_SFTP_NOT_CONNECTED_ERROR_MESSAGE,
      )
    for entry in self._sftp.listdir_attr(str(remote_dir)):
      entry_path = remote_dir.joinpath(entry.filename)
      if stat.S_ISDIR(entry.st_mode):
        yield from self._get_file_paths_in_remote_directory(entry_path)
      else:
        yield entry_path

  def pull_remote_directory(self, remote_src_dir: str,
                            local_dest_dir: str) -> None:
    """Pulls all files within a remote directory to local directory recursively.

    Args:
      remote_src_dir: The path to the remote source directory.
      local_dest_dir: The path to the local destination directory.
    """
    for remote_filepath in self._get_file_paths_in_remote_directory(
        pathlib.PurePosixPath(remote_src_dir)):
      # Use Path.joinpath to avoid messing up the forward/backward slashes.
      local_filepath = pathlib.Path(local_dest_dir).joinpath(
          remote_filepath.relative_to(remote_src_dir))
      self.pull(str(remote_filepath), str(local_filepath))

  def rm_dir_or_error(self, remote_dir: str) -> None:
    """Recursively force remove a remote directory or else raise an error.

    On Windows, busy files cannot be deleted. This function makes it easy to
    notice that problem. Usually the solution is to reboot the machine.
    Note that if the directory doesn't exist then no error will be thrown.

    Args:
      remote_dir: string, the remote directory.

    Raises:
      IOError: If the command failed with an error message showing that the
        device is busy. This matches RmFile's IOError.
      SSHRemoteError: If the command failed (exit code other than 0), but
        reason for the failure wasn't clear.
    """
    command_results = self.rm_dir(remote_dir)
    if command_results.exit_code:
      error_string = (
          f'exit code: {command_results.exit_code}, '
          f'output {command_results.output}, error: {command_results.error}')
      if 'Device or resource busy' in command_results.error:
        raise IOError('Likely a process is still using the file that '
                      f'you are trying to delete: {error_string}')
      else:
        raise SSHRemoteError(self, error_string)

  def rm_dir(self, remote_dir: str) -> CommandResults:
    """Recursively removes a remote directory from the remote machine.

    Args:
      remote_dir: the remote directory.

    Returns:
      A container with full command results.
    """
    if self.exists(remote_dir):
      command_results = CommandResults()
      self.execute_command(
          f'rm -rf "{remote_dir}"',
          ignore_error=True,
          command_results_collector=command_results)
      return command_results

    self.log.debug('Remote folder to remove %s does not exist', remote_dir)
    return CommandResults(exit_code=-1)

  def rm_file(self, remote_file: str) -> None:
    """Removes a remote file from the remote machine.

    Args:
      remote_file: the remote file location.

    Raises:
      SSHNotConnectedError: If sftp is not connected.
    """
    if self.exists(remote_file):
      if self._sftp is None:
        raise SSHNotConnectedError(
            ssh=self,
            message=_SFTP_NOT_CONNECTED_ERROR_MESSAGE,
        )
      self._sftp.remove(remote_file)
    else:
      self.log.debug('Remote file to remove %s does not exist', remote_file)

  def make_dirs(self, remote_dir: str) -> None:
    """Recursively makes directories on the remote machine.

    Args:
      remote_dir: string, the remote directory.

    Raises:
      RuntimeError: a component in remote_dir is not a directory.
      SSHNotConnectedError: If sftp is not connected.
    """
    if remote_dir != '/':
      remote_dir = remote_dir.rstrip('/')
    remote_stat_info = self.stat(remote_dir)
    if remote_stat_info:  # path exists
      if stat.S_ISDIR(remote_stat_info.st_mode):
        return  # already a directory; nothing to do.
      else:
        raise RuntimeError('%s is not a directory.' % remote_dir)

    # The 'or' below ensures that dirname of remote_dir is never empty
    self.make_dirs(os.path.dirname(remote_dir) or '.')
    if self._sftp is None:
      raise SSHNotConnectedError(
          ssh=self,
          message=_SFTP_NOT_CONNECTED_ERROR_MESSAGE,
      )
    self._sftp.mkdir(remote_dir)

  def is_dir(self, dir_path: str) -> bool:
    """Checks whether the dir_path is a directory.

    Args:
      dir_path: string, the path to the directory.

    Returns:
      True if dir_path is a valid path to a directory, False otherwise.
    """
    if dir_path != '/':
      dir_path = dir_path.rstrip('/')
    dir_stat_info = self.stat(dir_path)
    return dir_stat_info and stat.S_ISDIR(dir_stat_info.st_mode)  # pytype: disable=bad-return-type

  def is_file(self, file_path: str) -> bool:
    """Checks whether the file_path is a file.

    Args:
      file_path: string, a path to the file

    Returns:
      True if file_path is a valid path to a file, False otherwise.
    """
    file_path = file_path.rstrip('/')
    file_stat_info = self.stat(file_path)
    return file_stat_info and stat.S_ISREG(file_stat_info.st_mode)  # pytype: disable=bad-return-type

  def exists(self, remote_path: str) -> bool:
    """Checks whether the path exists on the remote machine.

    The path can be either to a file or to a folder.

    Args:
      remote_path: The path to the file on the remote machine.

    Returns:
      True if the file exists else False.
    """
    return self.stat(remote_path) is not None

  def stat(self, remote_path: str) -> Optional[sftp_attr.SFTPAttributes]:
    """Obtains stat info of a remote SFTP path on client.

    Args:
      remote_path: The path to the file on the remote machine.

    Returns:
      the SFTP attributes of the remote path.
    """
    try:
      if self._sftp is None:
        raise SSHNotConnectedError(
            ssh=self,
            message=_SFTP_NOT_CONNECTED_ERROR_MESSAGE,
        )
      return self._sftp.stat(remote_path)
    except IOError as e:
      # ENOENT means "no such file or directory."
      if e.errno == errno.ENOENT:
        return None
      raise

  def list_dir(self, path: str) -> Iterable[str]:
    """Return a list containing the names of the entries in the given path.

    The list is in arbitrary order. It does not include the special entries '.'
    and '..' even if they are present in the folder. This method is meant to
    mirror os.listdir as closely as possible.

    Args:
      path: path to list the contents of.

    Returns:
      List of strings contains the names of the entries.

    Raises:
      SSHNotConnectedError: If sftp is not connected.
    """
    if self._sftp is None:
      raise SSHNotConnectedError(
          ssh=self,
          message=_SFTP_NOT_CONNECTED_ERROR_MESSAGE,
      )
    return self._sftp.listdir(path)

  def chmod(self, remote_path: str, mode: int) -> None:
    """Changes the mode (permissions) of a remote file, similar to os.chmod.

    Args:
      remote_path: The path to the file on the remote machine.
      mode: The permission to set on the remote file.

    Raises:
      RemotePathDoesNotExistError: If the file to chmod is missing.
      SSHNotConnectedError: If sftp is not connected.
    """
    if not self.exists(remote_path):
      raise RemotePathDoesNotExistError(
          self,
          f'The file "{remote_path}" to change the permissions does not exist.')

    self.log.debug('chmod %o %r', mode, remote_path)
    if self._sftp is None:
      raise SSHNotConnectedError(
          ssh=self,
          message=_SFTP_NOT_CONNECTED_ERROR_MESSAGE,
      )
    self._sftp.chmod(remote_path, mode)

  def execute_command(
      self,
      command: str,
      timeout: Optional[float] = None,
      ignore_error: bool = False,
      command_results_collector: Optional[CommandResults] = None) -> str:
    """Executes the command in the remote machine.

    The command waits for the remote process to complete and returns the
    result. If the command you run produces large amount of output (either to
    stdout or to stderr), this method will cause a deadlock. For such commands
    please use execute_command_async() or start_remote_process().

    Args:
      command: The command to be executed in the remote machine.
      timeout: Optional timeout in seconds. If a non-negative float is given,
        subsequent channel read/write operations will raise a timeout exception
        if the timeout period value has elapsed before the operation has
        completed. Setting a timeout of None disables timeouts on socket
        operations.
      ignore_error: Whether to raise an exception if the command fails remotely.
      command_results_collector: the container to collect full command results.

    Returns:
      A string representing stripped stdout of the command.

    Raises:
      ExecuteCommandError: Raised if not ignoring error and exit code of command
      is greater than 0.
    """
    self.log.debug('Command to execute on remote machine: %s', command)

    channel_ = self.ssh_client.get_transport().open_session()
    if timeout is not None:
      channel_.settimeout(timeout)

    channel_.exec_command(command)

    command_result = _block_and_get_channel_status(channel_)
    self.log.debug(
        'cmd: %s, stdout: %s, stderr: %s, ret: %d',
        command,
        command_result.output,
        command_result.error,
        command_result.exit_code,
    )

    if command_results_collector:
      for key, value in dataclasses.asdict(command_result).items():
        setattr(command_results_collector, key, value)
    if not ignore_error and command_result.exit_code:
      raise ExecuteCommandError(self, command, command_result)

    return command_result.output.strip()

  def execute_command_async(
      self,
      command: str,
      timeout: Optional[int] = None,
      get_pty: bool = False
  ) -> tuple[channel.ChannelStdinFile, channel.ChannelFile,
             channel.ChannelStderrFile]:
    """Executes the command in the remote machine.

    This method returns immediately. It doesn't wait for the remote process to
    complete. Also, the remote process started using this command can outlive
    the connection to it. If you execute a long-running command and then exit
    python, then the command will still be running after you exit.

    Args:
      command: The command to be executed in the remote machine.
      timeout: The time in seconds to wait for the command to finish.
      get_pty: Whether to request a pseudo-terminal from the server. This is
        usually used right after creating a client channel, to ask the server to
        provide some basic terminal semantics for a shell invoked with
        invoke_shell. It is not necessary (or desirable) to enable it if you are
        going to execute a single command. But you need to enable it if you
        would like to get output stream continuously when the command is running
        even for a single command.

    Returns:
      The stdin, stdout, and stderr of the executing command.
    """
    self.log.debug('Running command on remote machine: %s', command)
    return self.ssh_client.exec_command(
        command, timeout=timeout, get_pty=get_pty)

  def execute_commands(self,
                       commands: Iterable[str],
                       sleep_interval: int = 0) -> None:
    """Executes batch of commands.

    Args:
      commands: A list of commands to be executed in the remote machine.
      sleep_interval: Number of seconds to sleep between commands.
    """
    for command in commands:
      self.ssh_client.exec_command(command)
      if sleep_interval:
        time.sleep(sleep_interval)

  def start_remote_process(
      self,
      command: str,
      environment: Optional[Mapping[str, str]] = None,
      timeout: Optional[int] = None,
      get_pty: bool = False,
      output_file_path: Optional[str] = None) -> RemotePopen:
    """Executes a command and returns a RemotePopen object.

    Args:
      command: The command to execute on the remote machine.
      environment: A dictionary of environment variables to enable beforehand.
      timeout: The time in seconds to wait for the command to finish.
      get_pty: Whether to request a pseudo-terminal from the server and execute
        the command using the pseudo-terminal. Turn on this switch to
        automatically exit the remote process when disconnecting the SSH client.
        Because the SSH client disconnection will kill the processes running in
        the pseudo-terminal.
      output_file_path: If set, the stdout and stderr of the remote process will
        be streamed to the file at the given host path.

    Returns:
      A RemotePopen to handle the process.
    """
    session = self.ssh_client.get_transport().open_session()
    if get_pty:
      session.get_pty()
    self.log.debug('Starting remote process: %s, environment=%s', command,
                   environment)
    return RemotePopen(
        self,
        session,
        command,
        environment,
        timeout,
        output_file_path=output_file_path)

  def get_remote_file_contents(self, remote_file: str) -> str:
    """Gets the content of a remote file.

    Args:
      remote_file: The remote file location.

    Returns:
      The content of a remote file.

    Raises:
      RemotePathDoesNotExistError: If the file is missing.
      SSHRemoteError: If the command failed (exit code other than 0).
    """
    if not self.exists(remote_file):
      raise RemotePathDoesNotExistError(
          self,
          f'The file to get the contents of ({remote_file}) does not exist on '
          'the remote machine.')

    try:
      stdout = self.execute_command(f'cat "{remote_file}"')
    except ExecuteCommandError as e:
      raise SSHRemoteError(self,
                           f'Error getting remote file: {remote_file}.') from e

    return stdout


class RemotePopen:
  """Manages a remote process.

  Uses the functionality of SSHProxy to build a Popen-like obj. Popen runs a
  process on the local machine, but in contrast this code runs the
  given command on a remote machine using ssh.SSHProxy.

  Attributes:
    pid: the remote process id on the remote machine.
  """

  def __init__(self,
               client: SSHProxy,
               session: channel.Channel,
               command: str,
               environment: Optional[Mapping[str, str]] = None,
               timeout: Optional[int] = None,
               output_file_path: Optional[str] = None) -> None:
    """Initializes the RemotePopen instance.

    Args:
      client: SSHProxy object with an open connection.
      session: SSH session to use for command execution
      command: The command to execute on the remote machine.
      environment: A dictionary of environment variables to enable beforehand.
      timeout: The time in seconds to wait for the command to finish.
      output_file_path: If set, the stdout and stderr of the remote process will
        be streamed to the file at the given host path.
    """
    if environment:
      mapping = ' '.join([f'{k}={v}' for k, v in environment.items()])
      set_env = f'export {mapping} && '
    else:
      set_env = ''
    if timeout is not None:
      session.settimeout(timeout)

    self._wait_timeout_sec = timeout

    if output_file_path is not None:
      # Combine stderr to stdout so we only need to stream the output of one
      # channel file
      session.set_combine_stderr(combine=True)

    # echo $$ writes the pid to stdout.
    session.exec_command(f'echo $$ && {set_env}exec {command}')

    self._client = client
    self._command = command
    self._session = session
    self._stdin = session.makefile('wb')
    self._stdout = session.makefile('rb')
    self._stderr = session.makefile_stderr('rb')
    self.pid = int(self._stdout.readline())

    self._client.log.debug('The remote process(pid=%d) started.', self.pid)

    self._output_streamer: Optional[
        channel_file_streamer.ChannelFileStreamer] = None
    if output_file_path is not None:
      self._start_streaming_remote_proc_output(output_file_path)

  def _start_streaming_remote_proc_output(self, output_file_path: str) -> None:
    """Starts streaming the remote process output to host."""
    logger = mobly_logger.PrefixLoggerAdapter(
        self._client.log,
        {
            mobly_logger.PrefixLoggerAdapter.EXTRA_KEY_LOG_PREFIX:
                f'[RemoteProcess({self.pid})]'
        },
    )
    self._output_streamer = channel_file_streamer.ChannelFileStreamer(
        self._stdout, output_file_path, logger)
    self._output_streamer.start()

  def communicate(self) -> tuple[str, str]:
    """Returns the stdout and stderr when the command finishes.

    If this object is streaming the remote process output, this method will
    block until the remote process exits or a timeout occurs, then return a pair
    of empty strings.

    Note that the process does not die when it times out. You need to kill it
    with the kill() method below.

    Returns:
      Tuple of stdout and stderr strings.

    Raises:
      RemoteTimeoutError: if the command did not complete in its given time.
    """
    if self._output_streamer is not None:
      if not self._wait_for_remote_process_exit(timeout=self._wait_timeout_sec):
        raise RemoteTimeoutError(
            self._client,
            f'Command did not complete in {self._wait_timeout_sec} seconds.',
        )
      self._stop_streaming_remote_proc_output()
      # Keep similar behavior to Popen, returning empty value when this class is
      # streaming output
      return ('', '')

    try:
      return (self._stdout.read().decode('utf-8'),
              self._stderr.read().decode('utf-8'))
    except socket.timeout as e:
      # socket.timeout is thrown during a read() operation when the timeout
      # given to the execute_command() function is exceeded.
      raise RemoteTimeoutError(
          self._client,
          f'Command did not complete in {self._session.gettimeout()} seconds.'
      ) from e

  def terminate(
      self,
      timeout: int | None = _SIGNAL_CMD_TIMEOUT_SEC,
      assert_process_exit: bool = False,
  ) -> None:
    """Attempts to terminate the remote process using SIGTERM.

    NOTE: If `assert_process_exit=True` and the process did not exit after
    sending SIGTERM, this will try force kill it with SIGKILL. If force kill
    still fails, this will raise an error.

    Args:
      timeout: The time limit in seconds to send the signal.
      assert_process_exit: If True, this will assert that the remote process
        exits after sending SIGTERM. If False, this will simply send the signal
        and return.

    Raises:
      SSHRemoteError: If `assert_process_exit=True` and the remote process did
        not exit.
    """
    self.send_signal(
        signal_id=signal.SIGTERM,
        timeout=timeout,
        assert_process_exit=assert_process_exit,
    )

  def poll(self) -> Optional[int]:
    """A non-blocking method to check the status of this process.

    Returns:
      The exit code of the process if the process has ended, otherwise None.
    """
    if not self._session.exit_status_ready():
      return None
    return self._session.recv_exit_status()

  def kill(
      self,
      timeout: int | None = _SIGNAL_CMD_TIMEOUT_SEC,
      assert_process_exit: bool = False,
  ) -> None:
    """Attempts to kill the remote process.

    NOTE: If `assert_process_exit=True` and the process did not exit after
    sending SIGKILL, this will raise an error.

    Args:
      timeout: The time limit in seconds to send the signal.
      assert_process_exit: If True, this will assert that the remote process
        exits after sending SIGKILL. If False, this will simply send the signal
        and return.

    Raises:
      SSHRemoteError: If `assert_process_exit=True` and the remote process did
        not exit.
    """
    self.send_signal(
        signal_id=signal.SIGKILL,
        timeout=timeout,
        assert_process_exit=assert_process_exit,
    )

  def send_signal(
      self,
      signal_id: int,
      timeout: int | None = _SIGNAL_CMD_TIMEOUT_SEC,
      assert_process_exit: bool = False,
  ) -> None:
    """Attempts to send a signal to the remote process.

    If the `output_file_path` argument is set in the constructor, then this
    method might need additional time to stop streaming output, which will not
    exceed `channel_file_streamer.CHANNEL_FILE_STREAMER_STOP_TIMEOUT_SEC`.

    This supports asserting the process exit after sending the signal.  If
    `assert_process_exit=True`, this will wait for process exit. If process
    did not exit after sending a non-SIGKILL signal, this will sending SIGKILL
    and assert the process exit.

    Args:
      signal_id: The ID of the signal. You can use constants defined in the
        "signal" module, e.g. signal.SIGTERM
      timeout: A timeout on blocking read/write operations. Default is 4 because
        we don't want to wait forever if this command fails.
      assert_process_exit: Whether to assert process exit after sending the
        signal.

    Raises:
      SSHRemoteError: If `assert_process_exit=True` and the process did not exit
        after sending SIGKILL.
    """
    self._do_send_signal(signal_id=signal_id, timeout=timeout)

    if not assert_process_exit:
      return

    if self._wait_for_remote_process_exit(timeout=_WAIT_PROC_EXIT_TIMEOUT_SEC):
      return

    if signal_id == signal.SIGKILL:
      raise SSHRemoteError(
          self._client, f'Failed to force kill the remote process {self.pid}.'
      )

    self._client.log.info(
        'Remote process pid=%d did not exit, sending escalated signal SIGKILL.',
        self.pid,
    )
    # It's ok that the force kill failed, we only need to check the remote
    # process exit status.
    with contextlib.suppress(
        RemoteTimeoutError, SSHRemoteError, ExecuteCommandError
    ):
      self.send_signal(
          signal_id=signal.SIGKILL, timeout=_SIGNAL_CMD_TIMEOUT_SEC
      )

    if not self._wait_for_remote_process_exit(
        timeout=_WAIT_PROC_EXIT_TIMEOUT_SEC
    ):
      raise SSHRemoteError(
          self._client,
          f'Failed to wait for remote process pid={self.pid} to exit.',
      )

  def _do_send_signal(self, signal_id: int, timeout: int | None = 4) -> None:
    """Sends a signal to the remote process."""
    if self.poll() is not None:
      self._stop_streaming_remote_proc_output()
      return

    command_results = CommandResults()
    try:
      self._client.execute_command(
          f'kill -{signal_id} {self.pid}',
          timeout=timeout,
          command_results_collector=command_results)
    except socket.timeout as e:
      raise RemoteTimeoutError(
          self._client,
          f'The kill command timed out after {self._wait_timeout_sec} seconds.'
      ) from e
    except socket.error as e:
      raise SSHRemoteError(
          self._client, 'Send signal failed due to socket error.'
      ) from e
    finally:
      self._stop_streaming_remote_proc_output()

    self._client.log.info(
        'Killed remote process command stdout: <%s>, stderr <%s>, exit code:'
        ' %s',
        command_results.output,
        command_results.error,
        self.poll(),
    )

  def wait(self,
           ignore_error: bool = False,
           command_results_collector: Optional[CommandResults] = None) -> str:
    """Waits until this process has ended and returns the command results.

    Args:
      ignore_error: Whether to raises an exception if the command fails
        remotely.
      command_results_collector: The container to collect full command results.

    Returns:
      A string representing stripped stdout of the command.

    Raises:
      ExecuteCommandError: Raised if not ignoring error and exit code of command
      was greater than 0.
      RemoteTimeoutError: Raised if the command did not complete in its given
      time.
    """
    command_results = command_results_collector or CommandResults()
    try:
      command_results.output, command_results.error = self.communicate()
    except RemoteTimeoutError:
      self.kill()
      raise
    finally:
      command_results.exit_code = self._session.recv_exit_status()

    self._client.log.debug('The remote process(pid=%d) ended.', self.pid)
    if not ignore_error and command_results.exit_code:
      raise ExecuteCommandError(self._client, self._command, command_results)

    self._client.log.debug('exit code: %s', command_results.exit_code)
    self._client.log.debug('output: %s', command_results.output)
    self._client.log.debug('error: %s', command_results.error)
    return command_results.output.strip()

  def _wait_for_remote_process_exit(self, timeout: int | None):
    """Waits for the remote process to exit."""
    deadline_time = None
    if timeout is not None:
      deadline_time = timeout + time.perf_counter()

    while not self._session.exit_status_ready():
      if (deadline_time is not None and time.perf_counter() > deadline_time):
        return False
      time.sleep(1)

    return True

  def _stop_streaming_remote_proc_output(self):
    """Stops streaming output of the remote process."""
    if self._output_streamer is not None:
      self._output_streamer.stop()


def _block_and_get_channel_status(channel_: channel.Channel) -> CommandResults:
  """Gets the exit code, stdout and stderr from a completed process.

  The channel will be blocked.

  Args:
    channel_: The channel to read status from.

  Returns:
    A container with full command results. Output and error are UTF-8 decoded.
  """
  # Blocking wait until the command completes.
  exit_code = channel_.recv_exit_status()

  stdout_file = channel_.makefile('rb')
  stderr_file = channel_.makefile_stderr('rb')

  with contextlib.closing(channel_):  # pytype: disable=wrong-arg-types
    with contextlib.closing(stdout_file):  # pytype: disable=wrong-arg-types
      stdout_str = stdout_file.read().decode('utf-8')
    with contextlib.closing(stderr_file):  # pytype: disable=wrong-arg-types
      stderr_str = stderr_file.read().decode('utf-8')

  return CommandResults(exit_code, stdout_str, stderr_str)


@contextlib.contextmanager
def create_ssh_connection(
    hostname: str,
    ssh_port: int = 22,
    username: str = 'root',
    password: Optional[str] = None,
    keyfile: Optional[str] = None,
    allow_agent: bool = False,
    connect_timeout: Optional[float] = None,
    connect_banner_timeout: Optional[float] = None,
):
  """The context manager to create a SSH connection."""
  ssh_connection = SSHProxy(
      hostname=hostname,
      ssh_port=ssh_port,
      username=username,
      password=password,
      keyfile=keyfile,
      allow_agent=allow_agent,
  )

  try:
    ssh_connection.connect(connect_timeout, connect_banner_timeout)
    yield ssh_connection
  finally:
    ssh_connection.disconnect()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configuration classes for the AP controller module."""

from collections.abc import Mapping, Sequence
import dataclasses
import enum
import random

import immutabledict
from mobly import utils

from mobly.controllers.wifi.lib import constants
from mobly.controllers.wifi.lib import errors
from mobly.controllers.wifi.lib.encryption import base_encryption_config
from mobly.controllers.wifi.lib.encryption import wpa


# TODO: move Ieee80211Standards from constants to this module.
Ieee80211Standards = constants.Ieee80211Standards

# For these channels, they support both HT40+ and HT40- so we cannot auto detect
# HT mode for them.
CHANNELS_HT40_PLUS_AND_MINUS = frozenset([5, 6, 7, 8, 9])

# Channels that only supports HT40+.
CHANNELS_HT40_PLUS = frozenset(
    [1, 2, 3, 4, 36, 44, 52, 60, 100, 108, 116, 124, 132, 140, 149, 157]
)

# Channels that only supports HT40-.
CHANNELS_HT40_MINUS = frozenset(
    [10, 11, 12, 13, 40, 48, 56, 64, 104, 112, 120, 128, 136, 144, 153, 161]
)

START_FREQ_FOR_WIDTH_80_SEGMENTS = (
    5180,
    5260,
    5500,
    5580,
    5660,
    5745,
    5955,
    6035,
    6115,
    6195,
    6275,
    6355,
    6435,
    6515,
    6595,
    6675,
    6755,
    6835,
    6915,
    6995,
)


@enum.unique
class ChannelWidth(enum.StrEnum):
  """The supported channel widths."""

  WIDTH_20 = '20MHz'
  WIDTH_40 = '40MHz'
  WIDTH_80 = '80MHz'

  def to_hostapd_enum(self) -> str:
    """Converts the channel width to a hostapd width enum."""
    match self:
      case ChannelWidth.WIDTH_20:
        return '0'
      case ChannelWidth.WIDTH_40:
        return '0'
      case ChannelWidth.WIDTH_80:
        return '1'


@enum.unique
class HTMode(enum.StrEnum):
  """High throughput mode."""

  NOHT = 'NOHT'
  HT20 = 'HT20'
  HT40_PLUS = 'HT40+'
  HT40_MINUS = 'HT40-'
  HT80 = 'HT80'


@enum.unique
class HostapdHTCapab(enum.StrEnum):
  """HT capabilities that can be set to `ht_capab` field of hostapd config."""

  HT40_PLUS = '[HT40+]'
  HT40_MINUS = '[HT40-]'
  SHORT_GI_20 = '[SHORT-GI-20]'
  SHORT_GI_40 = '[SHORT-GI-40]'
  LDPC = '[LDPC]'


@enum.unique
class HostapdVHTCapab(enum.StrEnum):
  """VHT capabilities that can be set to `vht_capab` field of hostapd config."""

  SHORT_GI_80 = '[SHORT-GI-80]'
  SHORT_GI_160 = '[SHORT-GI-160]'
  RXLDPC = '[RXLDPC]'


# The dict that returns whether a channel supports HT40+ or HT40-.
CHANNEL_HOSTAPD_HT40_MODE = immutabledict.immutabledict({
    **{ch: HostapdHTCapab.HT40_PLUS for ch in CHANNELS_HT40_PLUS},
    **{ch: HostapdHTCapab.HT40_MINUS for ch in CHANNELS_HT40_MINUS},
})


@enum.unique
class BandType(str, enum.Enum):
  """Band types."""

  BAND_2G = '2G'
  BAND_5G = '5G'


@enum.unique
class PMF(enum.IntEnum):
  """The enum for the setting of "Protected Management Frames" (IEEE802.11w)."""

  DISABLED = 0
  OPTIONAL = 1
  REQUIRED = 2


def generate_wifi_ssid(band_type: BandType) -> str:
  """Generates a wifi SSID with a randome substring."""
  random_str = utils.rand_ascii_str(5)
  return f'OpenWRT-{band_type.value}-{random_str}'


def generate_random_bssid() -> str:
  """Generates a random BSSID which is locally administered address and unicast.

  A LAA and unicast MAC address means the second digit must be 2, 6, a or
  e. I.e. X2:XX:XX:XX:XX:XX is valid, while X1:XX:XX:XX:XX:XX is not.

  Reference:
  https://en.wikipedia.org/wiki/MAC_address#Ranges_of_group_and_locally_administered_addresses

  Returns:
    Generated BSSID.
  """
  raw_values = [random.randrange(256) for _ in range(6)]
  raw_values[0] &= ~1
  raw_values[0] |= 2
  mac = ':'.join('%02x' % b for b in raw_values)
  return mac


def _transform_channel_width_to_ht_mode(
    width: ChannelWidth,
    channel: int,
    standard: Ieee80211Standards,
    ht_capab: Sequence[HostapdHTCapab],
) -> HTMode:
  """Transforms channel width to HT mode."""
  match width:
    case ChannelWidth.WIDTH_20:
      if standard in constants.STANDARDS_SUPPORT_HT_CAPAB:
        return HTMode.HT20
      else:
        return HTMode.NOHT

    case ChannelWidth.WIDTH_40:
      if channel in CHANNELS_HT40_PLUS_AND_MINUS:
        if HostapdHTCapab.HT40_PLUS in ht_capab:
          return HTMode.HT40_PLUS
        else:
          return HTMode.HT40_MINUS
      elif channel in CHANNELS_HT40_PLUS:
        return HTMode.HT40_PLUS
      else:
        return HTMode.HT40_MINUS

    case ChannelWidth.WIDTH_80:
      return HTMode.HT80


def _get_default_width(band_type: BandType) -> ChannelWidth:
  """Gets the default width with a given band type."""
  match band_type:
    case BandType.BAND_2G:
      return ChannelWidth.WIDTH_20
    case BandType.BAND_5G:
      return ChannelWidth.WIDTH_80


@dataclasses.dataclass
class WiFiConfig:
  """The class for what settings are required for the WiFi we need to start."""

  # Default to WiFi 6.
  standard: constants.Ieee80211Standards = constants.Ieee80211Standards.AC

  # If None, will be automatically generated in _post_init_.
  ssid: str | None = dataclasses.field(default=None, init=True)

  country_code: str = 'US'

  encryption_config: base_encryption_config.BaseEncryptionConfig = (
      dataclasses.field(default_factory=wpa.gen_config_for_wpa2_ccmp)
  )

  # See `constant.CHANNEL_TO_FREQUENCY` for all supported channels.
  channel: int = 1

  # The channel width.
  # If None, will be automatically decided in _post_init_.
  width: ChannelWidth | None = dataclasses.field(default=None, init=True)

  # HT capabilities (supported starting from 802.11N).
  # Corresponds to `ht_capab` field of hostapd config file.
  ht_capab: Sequence[HostapdHTCapab] | None = None

  # VHT capabilities (supported starting from 802.11AC).
  # Corresponds to `vht_capab` field of hostapd config file.
  vht_capab: Sequence[HostapdVHTCapab] | None = None

  # The setting of "Protected Management Frames" (IEEE802.11w).
  pmf: PMF | None = None

  # Whether to access wide area network (WAN) through network address
  # translation(NAT). If true, each wireless network will be on its own subnet
  # with its own dhcp server, and traffic will only be routed to specific subnet
  # when needed. Otherwise, all wireless networks will be shared together with
  # the WAN and it assumes there's a DHCP server running in the WAN.
  access_wan_through_nat: bool = True

  # Specifies the maximum desired transmission power in dBm. The actual txpower
  # used depends on regulatory requirements.
  # This must be a positive integer.
  # Reference value: by default 23 dBm will be used for channel 36 in US.
  maximum_txpower_dbm: int | None = None

  # If True, the Wi-Fi BSSID will be randomly generated. Otherwise the AP
  # will use its own MAC address as the Wi-Fi BSSID.
  use_random_bssid: bool = True

  custom_hostapd_configs: Mapping[str, str] = dataclasses.field(
      default_factory=dict
  )

  def __post_init__(self):
    if self.ssid is None:
      self.ssid = generate_wifi_ssid(self.band_type)
    if self.width is None:
      self.width = _get_default_width(self.band_type)

    self._check_validity()

  @property
  def band_type(self) -> BandType:
    """The band type."""
    if 1 <= self.channel <= 14:
      return BandType.BAND_2G
    elif 36 <= self.channel <= 177:
      return BandType.BAND_5G
    else:
      raise errors.ConfigError(
          f'Unsupported WiFi channel: {self.channel}. See'
          ' `constant.CHANNEL_TO_FREQUENCY` for all supported channels.'
      )

  def _check_validity(self):
    """Checks whether the configurations are valid."""
    self._check_frequency_configs_validity()
    self._check_txpower_config_validity()

  def _check_frequency_configs_validity(self):
    """Checks the frequency related configurations."""
    channel = self.channel
    if channel not in constants.CHANNEL_TO_FREQUENCY:
      raise errors.ConfigError(
          f'Unsupported WiFi channel: {channel}. See'
          ' `constant.CHANNEL_TO_FREQUENCY` for all supported channels.'
      )

    # Check the validity of HT40 mode when width >= 40MHz.
    ht_capab = self.ht_capab or []
    if self.width != ChannelWidth.WIDTH_20:
      if channel in CHANNELS_HT40_PLUS_AND_MINUS:
        if not (
            HostapdHTCapab.HT40_PLUS in ht_capab
            or HostapdHTCapab.HT40_MINUS in ht_capab
        ):
          raise errors.ConfigError(
              f'Must specify ht_capab HT40+/HT40- with channel={channel} and '
              f'width={self.width}.'
          )

      if (
          channel in CHANNELS_HT40_MINUS
          and HostapdHTCapab.HT40_PLUS in ht_capab
      ):
        raise errors.ConfigError(
            f'Got wrong ht_capab {HostapdHTCapab.HT40_PLUS.value} with'
            f' channel={channel} and width={self.width}.'
        )

      if (
          channel in CHANNELS_HT40_PLUS
          and HostapdHTCapab.HT40_MINUS in ht_capab
      ):
        raise errors.ConfigError(
            f'Got wrong ht_capab {HostapdHTCapab.HT40_MINUS.value} with'
            f' channel={channel} and width={self.width}.'
        )

    # Check the validity of combination (width, standard).
    match (self.width, self.standard):
      case (ChannelWidth.WIDTH_20, _):
        pass
      case (
          ChannelWidth.WIDTH_40,
          (
              Ieee80211Standards.N
              | Ieee80211Standards.AC
              | Ieee80211Standards.AX
          ),
      ):
        pass
      case (
          ChannelWidth.WIDTH_80,
          (Ieee80211Standards.AC | Ieee80211Standards.AX),
      ):
        pass
      case _:
        raise errors.ConfigError(
            f'Got unsupported channel width {self.width} with standard'
            f' {self.standard}.'
        )

  def _check_txpower_config_validity(self):
    """Checks the txpower configurations."""
    if self.maximum_txpower_dbm is not None and self.maximum_txpower_dbm <= 0:
      raise errors.ConfigError(
          'maximum_txpower_dbm must be positive integers, got'
          f' {self.maximum_txpower_dbm}.'
      )


@dataclasses.dataclass
class WifiInfo:
  """The class for the information of a WiFi running on the AP device."""

  id: int
  ssid: str
  password: str | None
  interface: str
  phy_name: str


@dataclasses.dataclass
class FreqConfig:
  """The class for representing the frequency band used by a wireless network.

  This class is using the same way to represent the frequency band as hostapd.
  See the docstring of `vht_oper_centr_freq_seg1_idx` for more details.

  When the band width is smaller than 80MHz, the frequency of the center1 should
  be always be zero. When the band width is 80MHz, the frequency of the center1
  will be automatically calculated.

  Attributes:
    channel: The control channel of the frequency band.
    ht_mode: The HT mode of the frequency band.
    center1_freq: The frequency of the center1.
  """

  channel: int
  ht_mode: HTMode
  center1_freq: int = 0

  def __post_init__(self):
    self._check_center_freq()

  def _check_center_freq(self):
    match self.ht_mode:
      case HTMode.NOHT | HTMode.HT20 | HTMode.HT40_PLUS | HTMode.HT40_MINUS:
        if self.center1_freq != 0:
          raise errors.ConfigError(
              'Specifying center1 frequency is not allowed for width'
              f' {self.ht_mode}.'
          )
      case HTMode.HT80:
        if self.center1_freq == 0:
          self.center1_freq = self._calc_center1_freq_for_width80()

  def _calc_center1_freq_for_width80(self) -> int:
    control_freq = constants.CHANNEL_TO_FREQUENCY[self.channel]
    for f in START_FREQ_FOR_WIDTH_80_SEGMENTS:
      if control_freq >= f and control_freq < f + 80:
        return f + 30
    raise errors.ConfigError(
        f'Got unsupported control frequency {control_freq} for width 80MHz.'
    )


def get_freq_config(wifi_config: WiFiConfig) -> FreqConfig:
  """Gets the frequency configuration from the given WiFi configuration."""
  ht_mode = _transform_channel_width_to_ht_mode(
      wifi_config.width,
      wifi_config.channel,
      wifi_config.standard,
      wifi_config.ht_capab or [],
  )
  return FreqConfig(channel=wifi_config.channel, ht_mode=ht_mode)


@dataclasses.dataclass(frozen=True)
class PcapConfig:
  """Configurations for controlling the packet capture process.

  Attributes:
    keep_latest_packets: True to ignore old packets if the capture file exceeds
      a default size limit. False to ignore new packets.
    ignore_qos_data_frames: Whether to ignore QoS data frames. Note that an
      exception is that this will not ignore EAPOL frames which are used for
      WPA2-PSK authentication.
  """

  keep_latest_packets: bool = True
  ignore_qos_data_frames: bool = False



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configurations for the encryption mode WPA."""

from collections.abc import Set
import enum

from mobly import utils as mobly_utils

from mobly.controllers.wifi.lib import errors
from mobly.controllers.wifi.lib import utils
from mobly.controllers.wifi.lib.encryption import base_encryption_config

_MIN_ASCII_PASSWORD_LEN = 8
_MAX_ASCII_PASSWORD_LEN = 63
_HEX_PASSWORD_LEN = 64


@enum.unique
class Cipher(enum.StrEnum):
  """Enum for WPA cipher suites for pairwise keys."""

  TKIP = 'TKIP'  # Temporal Key Integrity Protocol
  CCMP = 'CCMP'  # AES in Counter mode with CBC-MAC (CCMP-128)


@enum.unique
class KeyMgmt(enum.StrEnum):
  """Enum for WPA key management algorithms."""

  WPA_PSK = 'WPA-PSK'  # WPA-Personal / WPA2-Personal
  WPA_PSK_SHA256 = 'WPA-PSK-SHA256'  # WPA2-Personal using SHA256
  SAE = 'SAE'  # WPA3-Personal


@enum.unique
class Mode(enum.Enum):
  """Enum for WPA modes."""

  PURE_WPA = 'PureWPA'
  PURE_WPA2 = 'PureWPA2'
  PURE_WPA3 = 'PureWPA3'
  MIXED = 'WPAWPA2'
  MIXED_WPA3 = 'WPA2WPA3'


_MODES_USED_WPA = (Mode.PURE_WPA, Mode.MIXED)
_MODES_USED_WPA2 = (Mode.PURE_WPA2, Mode.PURE_WPA3, Mode.MIXED, Mode.MIXED_WPA3)
_MODES_USED_WPA3 = (Mode.PURE_WPA3, Mode.MIXED_WPA3)


def _generate_wifi_password():
  """Generates a wifi password with a random substring."""
  random_str = mobly_utils.rand_ascii_str(5)
  return f'RandomPSK-{random_str}'


class Wpa(base_encryption_config.BaseEncryptionConfig):
  """The configuration class for the WPA encryption type."""

  _mode: Mode
  _password: str
  _key_mgmt: Set[str]
  _ciphers: Set[str]  # ciphers used for WPA
  _ciphers2: Set[str]  # ciphers used for WPA2

  def __init__(
      self,
      mode: Mode = Mode.PURE_WPA2,
      password: str | None = None,
      ciphers: Set[Cipher] | None = None,
      ciphers2: Set[Cipher] | None = None,
      key_mgmt: Set[KeyMgmt] | None = None,
  ):
    """Constructor.

    Args:
      mode: The WPA mode to set.
      password: The password for the WiFi network. By default, a random password
        will be set.
      ciphers: A set of cipher suites (encryption algorithms) for pairwise keys
        (unicast packets) for WPA.
      ciphers2: A set of cipher suites (encryption algorithms) for pairwise keys
        (unicast packets) for RSN/WPA2.
      key_mgmt: A set of key management algorithms to be used.
    """
    self._mode = mode
    self._password = password or _generate_wifi_password()
    self._ciphers = ciphers or frozenset()
    self._ciphers2 = ciphers2 or frozenset()
    self._key_mgmt = key_mgmt or frozenset()
    self._validate()

  def _validate(self):
    """Validates the configurations are valid."""
    if self._mode not in _MODES_USED_WPA and self._ciphers:
      raise errors.ConfigError(
          f'ciphers cannot be specified for the mode {self._mode} that is not'
          ' using WPA1.'
      )

    if self._mode not in _MODES_USED_WPA2 and self._ciphers2:
      raise errors.ConfigError(
          f'ciphers2 cannot be specified for the mode {self._mode} that is not'
          ' using WPA2/RSN.'
      )

    if len(self._password) == _HEX_PASSWORD_LEN:
      # Password is in the form of hex string.
      if not utils.is_hex_string(self._password):
        raise errors.ConfigError(
            f'The password of length {_HEX_PASSWORD_LEN} must be a valid hex'
            f' string, got {self._password}'
        )
    else:
      # Password is in the form of ascii string.
      if not (
          _MIN_ASCII_PASSWORD_LEN
          <= len(self._password)
          <= _MAX_ASCII_PASSWORD_LEN
      ):
        raise errors.ConfigError(
            'The length of password must be in range'
            f' [{_MIN_ASCII_PASSWORD_LEN}, {_MAX_ASCII_PASSWORD_LEN}], got'
            f' {len(self._password)}'
        )

      if not self._password.isascii():
        raise errors.ConfigError(
            f'The password of length {len(self._password)} must be a valid'
            f' ascii string, got {self._password}'
        )

  def update_hostapd_conf(self, hostapd_conf):
    """See docstring of base class.

    Args:
      hostapd_conf: The hostapd config object to update.

    Raises:
      errors.ConfigError: If given hostapd configuration is not compatible with
        this encryption configuration.
    """
    mode_raw = 0
    if self._mode in _MODES_USED_WPA:
      mode_raw |= 1
    if self._mode in _MODES_USED_WPA2:
      mode_raw |= 2
    hostapd_conf.update('wpa', str(mode_raw))

    key_mgmt = self._key_mgmt
    if not key_mgmt:
      key_mgmt = set()
      # Include the minimum required key management algorithms.
      if self._mode != Mode.PURE_WPA3:
        key_mgmt.add(KeyMgmt.WPA_PSK.value)
      if self._mode in _MODES_USED_WPA3:
        key_mgmt.add(KeyMgmt.SAE.value)
    hostapd_conf.update('wpa_key_mgmt', ' '.join(key_mgmt))

    # PMF can be 0 (disabled) for WPA-PSK.
    pmf_needed = 0
    if KeyMgmt.SAE in key_mgmt or KeyMgmt.WPA_PSK_SHA256 in key_mgmt:
      pmf_needed = 2
    # WFA recommends PMF to be set to 1 (Optional) for WPA2/WPA3 mixed mode.
    if KeyMgmt.WPA_PSK in key_mgmt and pmf_needed == 2:
      pmf_needed = 1

    # Check if the PMF is set to the minimum required value.
    if hostapd_conf.get('ieee80211w') is None:
      pmf = 0  # Disabled by default.
    else:
      pmf = int(hostapd_conf.get('ieee80211w'))
    if pmf < pmf_needed:
      raise errors.ConfigError(
          f'PMF must be set to at least {pmf_needed}, while it is set to {pmf}'
      )

    hostapd_conf.set_password(self._password)
    if len(self._password) == _HEX_PASSWORD_LEN:
      hostapd_conf.update('wpa_psk', self._password)
    else:
      hostapd_conf.update('wpa_passphrase', self._password)

    if self._ciphers:
      hostapd_conf.update('wpa_pairwise', ' '.join(self._ciphers))

    if self._ciphers2:
      hostapd_conf.update('rsn_pairwise', ' '.join(self._ciphers2))


def gen_config_for_wpa2_ccmp() -> Wpa:
  """Generates a WPA2 configuration object with default settings."""
  return Wpa(
      mode=Mode.PURE_WPA2,
      password=None,
      ciphers2={Cipher.CCMP},
      key_mgmt={KeyMgmt.WPA_PSK},
  )


def gen_config_for_wpa3_ccmp() -> Wpa:
  """Generates a WPA3-Personal configuration object with default settings."""
  return Wpa(
      mode=Mode.PURE_WPA3,
      password=None,
      ciphers2={Cipher.CCMP},
      key_mgmt={KeyMgmt.SAE},
  )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants for the AP controller module."""

import datetime
import enum
import re

import immutabledict

CMD_SHORT_TIMEOUT = datetime.timedelta(seconds=10)

# A short summary:
# * US: 1 - 11, 36 - 64, 100 - 144, 149 - 173 are valid. 52 - 144 are DFS
#   channels.
# * Japan: 1 - 14, 36 - 64, 100 - 140 are valid. 52 - 140 are DFS channels.
CHANNEL_TO_FREQUENCY = immutabledict.immutabledict({
    # 2G
    1: 2412,
    2: 2417,
    3: 2422,
    4: 2427,
    5: 2432,
    6: 2437,
    7: 2442,
    8: 2447,
    9: 2452,
    10: 2457,
    11: 2462,
    12: 2467,
    13: 2472,
    14: 2484,
    # 5G
    36: 5180,
    38: 5190,
    40: 5200,
    42: 5210,
    44: 5220,
    46: 5230,
    48: 5240,
    52: 5260,
    56: 5280,
    60: 5300,
    64: 5320,
    100: 5500,
    104: 5520,
    108: 5540,
    112: 5560,
    116: 5580,
    120: 5600,
    124: 5620,
    128: 5640,
    132: 5660,
    136: 5680,
    140: 5700,
    144: 5720,
    149: 5745,
    153: 5765,
    157: 5785,
    161: 5805,
    165: 5825,
    169: 5845,
    173: 5865,
})


@enum.unique
class Ieee80211Standards(enum.StrEnum):
  """The enum for IEEE 802.11 standards."""

  A = 'IEEE802.11a'
  B = 'IEEE802.11b'
  G = 'IEEE802.11g'
  N = 'IEEE802.11n'
  AC = 'IEEE802.11ac'
  AX = 'IEEE802.11ax'


STANDARDS_SUPPORT_HT_CAPAB = (
    Ieee80211Standards.N,
    Ieee80211Standards.AC,
    Ieee80211Standards.AX,
)

STANDARDS_SUPPORT_VHT_CAPAB = (
    Ieee80211Standards.AC,
    Ieee80211Standards.AX,
)

STANDARDS_SUPPORT_HE_CAPAB = (Ieee80211Standards.AX,)


@enum.unique
class ApModel(enum.StrEnum):
  """The enum for AP models.

  The values should equal to the `deviceName` in the file
  `/etc/cros/cros_openwrt_image_build_info.json`.
  """

  U6LITE = 'Ubiquiti UniFi 6 Lite'
  BPIR3 = 'Bananapi BPi-R3'


@enum.unique
class IptablesAction(enum.StrEnum):

  INSERT = '-I'
  DELETE = '-D'


@enum.unique
class Commands(enum.StrEnum):
  """Commands to be executed on AP devices."""

  IP_FLUSH = 'ip addr flush {iface}'
  IP_ADDR_ADD = (
      'ip addr add {server_ip}/{network_mask_len} dev {iface} broadcast'
      ' {broadcast_ip}'
  )
  IP_LINK_SHOW = 'ip link show {interface}'
  IP_LINK_UP = 'ip link set {interface} up'

  IW_PHY = 'iw phy'
  IW_PHY_INFO = 'iw phy {phy} info'
  IW_REG_SET = 'iw reg set {country_code}'
  IW_REG_GET = 'iw reg get'
  IW_PHY_REG_GET = 'iw phy {phy} reg get'
  IW_DEV = 'iw dev'
  IW_DEV_DEL = 'iw dev {interface} del'
  IW_DEV_ADD = 'iw phy {phy} interface add {interface} type managed'
  IW_DEV_ADD_MONITOR = 'iw phy {phy} interface add {interface} type monitor'
  IW_DEV_INFO = 'iw dev {interface} info'
  IW_DEV_STATION_DUMP = 'iw dev {interface} station dump'
  IW_DEV_STATION_GET = 'iw dev {interface} station get {station_mac_address}'
  IW_DEV_SET_FREQ = 'iw dev {interface} set freq {freq_args}'
  IW_DEV_SET_MAXIMUM_TXPOWER = (
      'iw dev {interface} set txpower limit {txpower_mbm}'
  )
  IW_DEV_SCAN = 'iw dev {interface} scan {frequencies} {ssids}'

  # Firewall rules related commands
  FIREWALL_ENABLE_IP_FORWARD = 'echo 1 > /proc/sys/net/ipv4/ip_forward'
  FIREWALL_ENABLE_NAT = (
      'iptables -t nat -I POSTROUTING -o {interface} -j MASQUERADE'
  )
  FIREWALL_FORWARD_TRAFFIC = (
      'iptables {action} FORWARD -i {in_interface} -o {out_interface} -j ACCEPT'
  )
  FIREWALL_FORWARD_KNOWN_TRAFFIC = (
      'iptables {action} FORWARD -i {in_interface} -o {out_interface} -m'
      ' conntrack --ctstate ESTABLISHED,RELATED -j ACCEPT'
  )

  START_TCPDUMP = (
      'tcpdump -vv -i {interface} -U -e -B 1024 -w {file_path} {args}'
  )

  # Opkg commands.
  OPKG_LIST = 'opkg list-installed {package}'
  OPKG_UPDATE = 'opkg update'
  OPKG_INSTALL = 'opkg install {package}'

  HOSTAPD_START = '/usr/sbin/hostapd -dd -t -K {conf_path}'

  KILLALL = 'killall {name}'

  GET_PROCESS_BY_NAME = 'ps | grep {name}'

  REBOOT = 'sudo reboot'

  CHECK_DEVICE_REBOOT_READY = 'test -f /tmp/cros/status/ready'

  # The command to obtain syslog in OpenWrt systems.
  LOGREAD = 'logread -f'


# The network interface to for the AP device to connect with wide area network.
WAN_INTERFACE = 'br-lan'

# Constant fot the name of hostapd.
HOSTAPD = 'hostapd'


# The AP device SSH username.
SSH_USERNAME = 'root'

REMOTE_WORK_DIR = '/tmp/mobly_artifacts'
OPENWRT_RELEASE_INFO_FILE_PATH = '/etc/openwrt_release'
CURSTOM_RELEASE_INFO_FILE_PATH = '/etc/cros/cros_openwrt_image_build_info.json'

OPENWRT_PACKAGE_IPTABLES = 'iptables'
OPENWRT_PACKAGE_SFTP = 'openssh-sftp-server'
OPENWRT_PACKAGE_SUDO = 'sudo'
OPENWRT_PACKAGE_HOSTAPD = 'hostapd'
OPENWRT_PACKAGE_TCPDUMP = 'tcpdump'
OPENWRT_PACKAGE_IPERF3 = 'iperf3'

# Required packages when using an OpenWrt image built against snapshot.
# * Do NOT include `iptables` because snapshot images have limitation on
#   installing kernel version dependent modules. Reference:
#   https://openwrt.org/releases/snapshot
REQUIRED_PACKAGES_SNAPSHOT_IMAGE = (
    OPENWRT_PACKAGE_SFTP,
    OPENWRT_PACKAGE_SUDO,
    OPENWRT_PACKAGE_TCPDUMP,
)

# Required packages when using an official OpenWrt image.
# * Do NOT include `hostapd` because it might be installed but not through
#   opkg.
REQUIRED_PACKAGES_RELEASED_IMAGE = (
    OPENWRT_PACKAGE_SFTP,
    OPENWRT_PACKAGE_SUDO,
    OPENWRT_PACKAGE_TCPDUMP,
)

# Required packages when using AP model BPi R3 and a CrOS customized image.
# The image used in BPi R3 is yet fully upstreamed thus installing package from
# online repos doesn't work. So we assume the packages are already installed.
REQUIRED_PACKAGES_BPIR3_AND_CROS_BUILT_IMAGE = tuple()

DEVICE_INFO_PATTERN = re.compile(
    r"DISTRIB_ID='(?P<image_type>.*)'\n"
    r"DISTRIB_RELEASE='(?P<release>.*)'\n"
    r"DISTRIB_REVISION='(?P<revision>.*)'\n"
    r"DISTRIB_TARGET='(?P<target>.*)'\n"
    r"DISTRIB_ARCH='(?P<arch>.*)'\n"
    r"DISTRIB_DESCRIPTION='(?P<description>.*)'\n"
    r"DISTRIB_TAINTS='(?P<taints>.*)'"
)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Controlling DHCP server instances on AP devices."""

from collections.abc import Set
import contextlib
import dataclasses
import datetime
import ipaddress
import logging
import os
import threading
import typing
from typing import Any

from mobly import logger as mobly_logger

from mobly.controllers.wifi.lib import ssh as ssh_lib
from mobly.controllers.wifi.lib import constants

OpenWrtDevice = Any


# We intentionally do not set the log-facility configuration because we can
# directly capture logs through remote process output.
_DNSMASQ_CONFIG = """
dhcp-range={network_ip_start},{network_ip_end}
interface={iface}
log-dhcp
no-resolv
no-hosts
server=8.8.8.8
server=8.8.4.4
log-queries
except-interface=lo
bind-interfaces
dhcp-leasefile={lease_file_remote_path}
"""

_REMOTE_CMD_DEFAULT_TIMEOUT = datetime.timedelta(seconds=30)


@dataclasses.dataclass()
class DhcpConfig:
  """Configurations for a DhcpManager instance."""

  # The ID of the subnet where we will operate a DHCP server instance.
  subnet_id: int
  # The wireless interface name.
  iface: str

  # The subnet information structure.
  _subnet: ipaddress.IPv4Network | None = dataclasses.field(
      default=None, init=False
  )

  @property
  def subnet(self) -> ipaddress.IPv4Network:
    """The subnet information structure."""
    if self._subnet is None:
      self._subnet = typing.cast(
          ipaddress.IPv4Network,
          ipaddress.ip_network(f'192.168.{self.subnet_id}.0/24'),
      )
    return self._subnet

  @property
  def broadcast_ip(self) -> str:
    """The broadcast ip address."""
    return str(self.subnet.broadcast_address)

  @property
  def server_ip(self) -> str:
    """The server ip address."""
    # Use -2 here because `self.subnet[-1]` is the broadcast ip address.
    return str(self.subnet[-2])

  @property
  def assignable_ip_range(self) -> tuple[str, str]:
    """The range of the IPs that can be assigned to clients."""
    return (str(self.subnet[0]), str(self.subnet[128]))

  @property
  def mask_len(self) -> int:
    """The length of the network mask."""
    return self.subnet.prefixlen


_SUBNET_ID_MAX_COUNT = 256


class _SubnetIdGenerator:
  """The class to generate subnet IDs."""

  _busy_ids: Set[int]
  _counter: int
  _lock: threading.Lock

  def __init__(self):
    self._busy_ids = set()
    self._counter = 0
    self._lock = threading.Lock()

  def get(self) -> int:
    """Gets a free subnet ID."""
    with self._lock:
      for i in range(_SUBNET_ID_MAX_COUNT):
        counter = (self._counter + i) % _SUBNET_ID_MAX_COUNT
        if counter not in self._busy_ids:
          self._busy_ids.add(counter)
          self._counter = (counter + 1) % _SUBNET_ID_MAX_COUNT
          return counter

    raise RuntimeError(
        'Cannot find a free subnet ID because all'
        f' {_SUBNET_ID_MAX_COUNT} IDs are in use.'
    )

  def release(self, subnet_id: int) -> None:
    """Releases a subnet ID so it can be used by other subnets.

    This is a no-op if the provided ID is not in use.

    Args:
      subnet_id: The subnet ID to release.
    """
    with self._lock:
      if subnet_id in self._busy_ids:
        self._busy_ids.remove(subnet_id)


_subnet_id_generator = _SubnetIdGenerator()


class DhcpManager:
  """The class that manages one DHCP server instance on the AP device."""

  def __init__(
      self,
      device: 'OpenWrtDevice',
      wifi_id: int,
      iface: str,
      base_logger: (
          logging.Logger | mobly_logger.PrefixLoggerAdapter | None
      ) = None,
  ):
    """Constructor.

    Args:
      device: The AP device controller object.
      wifi_id: The ID of the WiFi network where we will operate a DHCP server
        instance.
      iface: The wireless interface name.
      base_logger: The base logger. Based on that logger, this class will prefix
        each log entry with string "[DhcpManager]".
    """
    self._device = device
    self._wifi_id = wifi_id
    self._subnet_id = _subnet_id_generator.get()
    self._remote_process = None
    base_logger = base_logger or device.log
    self._log = mobly_logger.PrefixLoggerAdapter(
        base_logger,
        {
            mobly_logger.PrefixLoggerAdapter.EXTRA_KEY_LOG_PREFIX: (
                f'[DhcpManager|wifi{self._wifi_id}]'
            )
        },
    )
    self._config = DhcpConfig(subnet_id=self._subnet_id, iface=iface)
    self._identifier = f'wifi{self._wifi_id},{self._config.iface}'

    self._local_work_dir = self._device.log_path
    self._remote_work_dir = self._device.remote_work_dir
    self._artifact_time_str = mobly_logger.get_log_file_timestamp()

  def start(self) -> None:
    """Starts this DHCP manager."""
    self._log.debug(
        'Starting a remote DHCP server instance with config %s', self._config
    )
    try:
      self._start()
    except ssh_lib.SSHRemoteError:
      self._log.error('Stopping due to start failure.')
      with contextlib.suppress(ssh_lib.SSHRemoteError):
        self.stop()
      raise

  def _start(self) -> None:
    """Starts a DHCP server instance on the AP device."""
    # Set the ip configurations to the wireless interface.
    self._device.ssh.execute_command(
        constants.Commands.IP_FLUSH.format(iface=self._config.iface),
        timeout=_REMOTE_CMD_DEFAULT_TIMEOUT.total_seconds(),
    )
    self._device.ssh.execute_command(
        constants.Commands.IP_ADDR_ADD.format(
            server_ip=self._config.server_ip,
            network_mask_len=self._config.mask_len,
            iface=self._config.iface,
            broadcast_ip=self._config.broadcast_ip,
        ),
        timeout=_REMOTE_CMD_DEFAULT_TIMEOUT.total_seconds(),
    )

    # Start a dnsmasq process on the remote device.
    config_file_remote_path = self._generate_remote_config_file()
    self._start_dnsmasq_process(config_file_remote_path)

  def _generate_remote_config_file(self):
    """Generates the configurations into a config file on the AP device."""
    # Generate the content of dnsmasq config.
    config_content = _DNSMASQ_CONFIG.format(
        network_ip_start=self._config.assignable_ip_range[0],
        network_ip_end=self._config.assignable_ip_range[1],
        iface=self._config.iface,
        lease_file_remote_path=self._get_remote_path(
            filename=self._get_lease_filename(),
        ),
    )

    # Write dnsmasq config to a remote file.
    filename = self._get_conf_filename()
    local_path = self._get_local_path(filename)
    remote_path = self._get_remote_path(filename)
    with open(local_path, 'w') as f:
      f.write(config_content)
    self._device.push_file(local_path, remote_path)
    # Rename the local file so it can be directly opened on Sponge.
    os.rename(local_path, f'{local_path}.txt')

    return remote_path

  def _start_dnsmasq_process(self, config_remote_path: str):
    local_path = self._get_local_path(filename=self._get_log_filename())
    command_elements = [
        'dnsmasq',
        f'--conf-file={config_remote_path}',
        '--no-daemon',
    ]
    command = ' '.join(command_elements)
    self._remote_process = self._device.ssh.start_remote_process(
        command, get_pty=True, output_file_path=local_path
    )
    self._log.debug(
        'Started dnsmasq remote process %d on the AP device.',
        self._remote_process.pid,
    )

  def __del__(self):
    self.stop()

  def stop(self):
    """Stops the DHCP server instance on the AP device."""
    if self._subnet_id is not None:
      _subnet_id_generator.release(self._subnet_id)
      self._subnet_id = None

    proc = self._remote_process
    self._remote_process = None

    if proc is None:
      return

    self._log.debug('Stopping remote DHCP server.')

    proc.terminate(
        timeout=_REMOTE_CMD_DEFAULT_TIMEOUT.total_seconds(),
        assert_process_exit=True,
    )
    self._log.debug(
        'Killed dnsmasq remote process %d with poll value: %s',
        proc.pid,
        proc.poll(),
    )

  def _get_remote_path(self, filename: str) -> str:
    return os.path.join(self._remote_work_dir, filename)

  def _get_local_path(self, filename: str) -> str:
    return os.path.join(self._local_work_dir, filename)

  def _get_conf_filename(self) -> str:
    return f'{self._identifier},dnsmasq.conf'

  def _get_log_filename(self) -> str:
    return f'{self._identifier},dnsmasq.log'

  def _get_lease_filename(self) -> str:
    return f'{self._identifier},dnsmasq.leases'



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configurations for the encryption mode OPEN."""

from mobly.controllers.wifi.lib.encryption import base_encryption_config


class Open(base_encryption_config.BaseEncryptionConfig):
  """The configuration class for the encryption type OPEN."""

  def update_hostapd_conf(self, hostapd_conf):
    hostapd_conf.set_password(None)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for the AP controller module."""

from collections.abc import Callable
import datetime
import string
import time
from typing import Any

from mobly import utils
from packaging import version

from mobly.controllers.wifi.lib import constants

OpenWrtDevice = Any

_OPENWRT_NEW_FIREWALL_RULE_VERSION = version.parse('22.03')


def is_hex_string(s: str) -> bool:
  """True if the given string is a hex string; False otherwise."""
  return all(c in string.hexdigits for c in s)


def wait_for_predicate(
    predicate: Callable[[], bool],
    timeout: datetime.timedelta,
    interval: datetime.timedelta | None = None,
) -> bool:
  """Returns True if the predicate returns True within the given timeout.

  Any exception raised in the predicate will terminate the wait immediately.

  Args:
    predicate: A predicate function.
    timeout: The timeout to wait.
    interval: The interval time between each check of the predicate.

  Returns:
    Whether the predicate returned True within the given timeout.
  """
  start_time = time.monotonic()
  deadline = start_time + timeout.total_seconds()
  while time.monotonic() < deadline:
    if predicate():
      return True
    if interval is not None:
      time.sleep(interval.total_seconds())
  return False


def is_new_firewall_rule_version(version_number: str) -> bool:
  """Returns True if OpenWrt version is new firewall rule, False otherwise."""
  parsed_version = version.parse(version_number)
  return parsed_version >= _OPENWRT_NEW_FIREWALL_RULE_VERSION


def is_using_openwrt_snapshot_image(release: str) -> bool:
  """Returns True if the image is built against SNAPSHOT, False otherwise."""
  return release == 'SNAPSHOT'


def is_using_custom_image(device: 'OpenWrtDevice') -> bool:
  """Returns True if the image is using a custom image, False otherwise."""
  output = device.ssh.execute_command(
      command=f'ls {constants.CURSTOM_RELEASE_INFO_FILE_PATH}',
      timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
      ignore_error=True,
  )
  return output.strip() == constants.CURSTOM_RELEASE_INFO_FILE_PATH


def run_command(
    cmd: str, ignore_error: bool = False
) -> tuple[int, bytes, bytes]:
  """Runs a command in a subprocess.

  Args:
    cmd: The command to run.
    ignore_error: If False, raise an error if the command execution failed.

  Returns:
    A 3-tuple of the consisting of the return code, the std output, and the
      std error.

  Raises:
    RuntimeError: The command execution failed and `ignore_error=False`.
  """
  ret, out, err = utils.run_command(cmd, shell=True)
  if (not ignore_error) and ret != 0:
    raise RuntimeError(f'Failed to run command "{cmd}" with error: {err}')
  return ret, out, err


def convert_testbed_bool_value(value: bool | str) -> bool:
  """Converts a raw value from testbed configuration to a bool value.

  We need this method because in some trigger approaches bool values in MH
  static testbed are transformed to strings in Mobly testbed.

  Args:
    value: The raw value from testbed configuration.

  Returns:
    The bool value.

  Raises:
    ValueError: If got invalid value.
  """
  if isinstance(value, bool):
    return value
  if isinstance(value, str):
    if value.lower() == 'true':
      return True
    if value.lower() == 'false':
      return False
  raise ValueError(f'Invalid bool value from testbed: {value}')



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The module for controlling hostpad on AP devices.

The responsibility of this module mainly include:
* Generate hostapd configuration files on AP devices.
* Start/Stop hostpad processes on AP devices.
"""

from collections.abc import Sequence
import contextlib
import datetime
import logging
import os
import typing
from typing import Any

import immutabledict
from mobly import logger as mobly_logger

from mobly.controllers.wifi.lib import ssh as ssh_lib
from mobly.controllers.wifi.lib import constants
from mobly.controllers.wifi.lib import errors
from mobly.controllers.wifi.lib import iw_utils
from mobly.controllers.wifi.lib import utils
from mobly.controllers.wifi.lib import wifi_configs

OpenWrtDevice = Any


# The network bridge set up by OpenWrt to bridge wireless networks with the wide
# area network (WAN).
_NETWORK_BRIGE_CONNECTING_WAN = 'br-lan'


_WIFI_START_WAIT_TIME = datetime.timedelta(seconds=30)
# DFS channels require 60s extra start time to check channel availability.
_WIFI_START_WAIT_TIME_DFS = datetime.timedelta(seconds=180)
_WIFI_START_CHECK_INTERVAL = datetime.timedelta(seconds=5)
_WIFI_STOP_WAIT_TIME = datetime.timedelta(seconds=30)


_WIFI_CONFIG_TO_HW_MODE = immutabledict.immutabledict({
    constants.Ieee80211Standards.A: {
        wifi_configs.BandType.BAND_5G: 'a',
    },
    constants.Ieee80211Standards.B: {
        wifi_configs.BandType.BAND_2G: 'b',
    },
    constants.Ieee80211Standards.G: {
        wifi_configs.BandType.BAND_2G: 'g',
    },
    constants.Ieee80211Standards.N: {
        wifi_configs.BandType.BAND_2G: 'g',
        wifi_configs.BandType.BAND_5G: 'a',
    },
    constants.Ieee80211Standards.AC: {
        wifi_configs.BandType.BAND_2G: 'g',
        wifi_configs.BandType.BAND_5G: 'a',
    },
    constants.Ieee80211Standards.AX: {
        wifi_configs.BandType.BAND_2G: 'g',
        wifi_configs.BandType.BAND_5G: 'a',
    },
})

# These default capabilities are enabled to improve the throughput. This is a
# subset of the default settings of UCI.
_DEFAULT_HT_CAPAB = {
    wifi_configs.BandType.BAND_2G: (
        # Enable short guard interval.
        wifi_configs.HostapdHTCapab.SHORT_GI_20,
        wifi_configs.HostapdHTCapab.SHORT_GI_40,
    ),
    wifi_configs.BandType.BAND_5G: (
        # Enable short guard interval.
        wifi_configs.HostapdHTCapab.SHORT_GI_20,
        wifi_configs.HostapdHTCapab.SHORT_GI_40,
        # Enable LDPC codes for error correction. 5GHz only.
        wifi_configs.HostapdHTCapab.LDPC,
    ),
}
_DEFAULT_VHT_CAPAB = {
    wifi_configs.BandType.BAND_2G: (),
    wifi_configs.BandType.BAND_5G: (
        # Enable short guard interval.
        wifi_configs.HostapdVHTCapab.SHORT_GI_80,
        # Enable LDPC codes when receiving and decoding.
        wifi_configs.HostapdVHTCapab.RXLDPC,
    ),
}

# We set fragm_threshold to -1 because the AP might not support setting
# fragmentation threshold when it is not using our custom OpenWrt image. And I
# did not find a way to dynamically check whether setting fragmentation
# threshold is supported or not.
_IS_CUSTOM_OPENWRT_TO_DEFAULT_FRAGM_THRESHOLD = {
    True: 2346,
    False: -1,
}


def _get_default_ht_capab(
    wifi_config: wifi_configs.WiFiConfig,
) -> Sequence[str]:
  """Gets default `ht_capab` configuration."""
  return _DEFAULT_HT_CAPAB[wifi_config.band_type]


def _get_default_vht_capab(
    wifi_config: wifi_configs.WiFiConfig,
) -> Sequence[str]:
  """Gets default `vht_capab` configuration."""
  return _DEFAULT_VHT_CAPAB[wifi_config.band_type]


def _get_center_channel_with_width_80mhz(channel_20mhz: int) -> int:
  """Gets the center channel of a segment of width 80MHz.

  The AP will combines 4 20MHz channels into 1 80MHz. This method takes one
  20MHz channel and returns the 80MHz channel that covers it.

  See Figure "5 GHz Channels, with DFS and TDWR Restrictions" in the following
  link for the correspondence between 20MHz channels and 80MHz channels:

  https://revolutionwifi.blogspot.com/2013/03/80211ac-channel-planning.html

  Args:
    channel_20mhz: A 20MHz channel that is covered by the 80MHz channel.

  Returns:
    The center channel of the 80MHz channel.
  """
  match channel_20mhz:
    case 36 | 40 | 44 | 48:
      return 42
    case 52 | 56 | 60 | 64:
      return 58
    case 100 | 104 | 108 | 112:
      return 106
    case 116 | 120 | 124 | 128:
      return 122
    case 132 | 136 | 140 | 144:
      return 138
    case 149 | 153 | 157 | 161:
      return 155
    case _:
      raise errors.ConfigError(
          'Got unsupported 20MHz channel when using channel width 80MHz:'
          f' {channel_20mhz}'
      )


class HostapdConfig:
  """The hostapd configurations.

  When we need to start a WiFi network, we need to transform the user specified
  configurations to a hostapd configuration file. This class representes the
  hostapd configuration file.

  Attributes:
    channel: The WiFi channel.
    ssid: The WiFi SSID.
    password: The WiFi password.
    interface: The name of the network interface that the WiFi network is using.
    config_content: The content of the hostapd configuration file.
  """

  channel: int
  ssid: str
  password: str | None
  interface: str

  _raw: dict[str, str]

  def __init__(
      self, interface: str, dfs_channels: set[int], is_custom_openwrt: bool
  ):
    self._dfs_channels = dfs_channels
    self._raw = {}
    self.update('logger_syslog', '-1')
    self.update('logger_syslog_level', '0')
    self.update('rts_threshold', '-1')
    self.update('driver', 'nl80211')
    self.update(
        'fragm_threshold',
        str(_IS_CUSTOM_OPENWRT_TO_DEFAULT_FRAGM_THRESHOLD[is_custom_openwrt]),
    )
    self.set_interface(interface)

  def update(self, key: str, value: str):
    self._raw[key] = value

  def get(self, key: str) -> str | None:
    """Gets the value of the given key.

    Args:
      key: The key to get the value.

    Returns:
      The value of the given key, or None if the key is not found.
    """
    if key not in self._raw:
      return None
    return self._raw[key]

  def update_from_wifi_config(self, wifi_config: wifi_configs.WiFiConfig):
    """Updates this object according to the `WiFiConfig` object."""
    self.set_ssid(wifi_config.ssid)
    if wifi_config.use_random_bssid:
      bssid = wifi_configs.generate_random_bssid()
      self.update('bssid', bssid)

    # TODO: Better API is returning a dict / hostapd_conf and merge
    # it.
    self._update_encryption_configs(wifi_config)
    self.set_channel(wifi_config.channel)
    self._update_dfs_channel_config(wifi_config)

    self.update('ieee80211d', '1')  # Required when country_code is set.
    self.update('country_code', wifi_config.country_code)
    self.update('hw_mode', self._get_hw_mode(wifi_config))
    self._update_according_to_wifi_standard(wifi_config)

    if not wifi_config.access_wan_through_nat:
      self.update('bridge', _NETWORK_BRIGE_CONNECTING_WAN)

    if wifi_config.pmf is not None:
      self.update('ieee80211w', str(wifi_config.pmf.value))

    for key, value in wifi_config.custom_hostapd_configs.items():
      self.update(key, value)

  def set_ssid(self, value: str):
    self.ssid = value
    self.update('ssid2', f'"{value}"')

  def set_password(self, value: str | None):
    self.password = value

  def set_interface(self, value: str):
    self.interface = value
    self.update('interface', value)

  def set_channel(self, value: int):
    self.channel = value
    self.update('channel', str(value))

  def is_using_a_dfs_channel(self) -> bool:
    """Returns true if the channel is a DFS channel."""
    return self.channel in self._dfs_channels

  @property
  def config_content(self) -> str:
    return '\n'.join(f'{key}={value}' for key, value in self._raw.items())

  def write_to_file(self, filepath: str, content: str) -> None:
    """Writes the configurations to the given host filepath."""
    with open(filepath, 'w') as f:
      f.write(content)

  def _get_hw_mode(self, wifi_config: wifi_configs.WiFiConfig) -> str:
    """Get the hwmode."""
    band_type_to_hw_mode = _WIFI_CONFIG_TO_HW_MODE.get(wifi_config.standard)
    if band_type_to_hw_mode is None:
      raise errors.ConfigError(
          f'Got unknown WiFi standard: {wifi_config.standard}'
      )

    hw_mode = band_type_to_hw_mode.get(wifi_config.band_type)
    if hw_mode is None:
      raise errors.ConfigError(
          f'Got unsupported band type "{wifi_config.band_type}" under WiFi'
          f' standard "{wifi_config.standard}"'
      )
    return hw_mode

  def _update_dfs_channel_config(self, wifi_config: wifi_configs.WiFiConfig):
    if not self.is_using_a_dfs_channel():
      return
    if wifi_config.country_code is None:
      raise errors.ConfigError(
          'Country code must be set when using a DFS channel.'
      )

    # This is required for ieee80211h
    self.update('ieee80211d', '1')
    # This enables radar detection and DFS support.
    self.update('ieee80211h', '1')

  def _update_according_to_wifi_standard(
      self, wifi_config: wifi_configs.WiFiConfig
  ):
    self._update_11n_configs(wifi_config)
    self._update_11ac_configs(wifi_config)
    self._update_11ax_configs(wifi_config)

  def _update_11n_configs(self, wifi_config: wifi_configs.WiFiConfig):
    """Updates 802.11N related configurations."""
    if wifi_config.standard not in constants.STANDARDS_SUPPORT_HT_CAPAB:
      return
    self.update('ieee80211n', '1')
    self.update('wmm_enabled', '1')  # Required when HT capabilities are used.
    if ht_capab := self._get_ht_capab(wifi_config):
      self.update('ht_capab', ht_capab)

  def _get_ht_capab(self, wifi_config: wifi_configs.WiFiConfig) -> str:
    """Gets the value for `ht_capab` in hostapd config file."""
    ht_capab = wifi_config.ht_capab
    if ht_capab is None:
      ht_capab = list(_get_default_ht_capab(wifi_config))
    if wifi_config.width != wifi_configs.ChannelWidth.WIDTH_20:
      capab = wifi_configs.CHANNEL_HOSTAPD_HT40_MODE.get(wifi_config.channel)
      if capab is not None and capab not in ht_capab:
        ht_capab = ht_capab + [capab]
    return ''.join(ht_capab)

  def _update_11ac_configs(self, wifi_config: wifi_configs.WiFiConfig):
    """Updates 802.11AC related configurations."""
    if wifi_config.standard not in constants.STANDARDS_SUPPORT_VHT_CAPAB:
      return
    self.update('ieee80211ac', '1')

    width = typing.cast(wifi_configs.ChannelWidth, wifi_config.width)
    self.update('vht_oper_chwidth', width.to_hostapd_enum())

    if wifi_config.band_type == wifi_configs.BandType.BAND_2G:
      return

    if wifi_config.width == wifi_configs.ChannelWidth.WIDTH_80:
      self.update(
          'vht_oper_centr_freq_seg0_idx',
          str(_get_center_channel_with_width_80mhz(wifi_config.channel)),
      )

    # vht_capab
    vht_capab = wifi_config.vht_capab
    if vht_capab is None:
      vht_capab = _get_default_vht_capab(wifi_config)
    if vht_capab:
      self.update('vht_capab', ''.join(vht_capab))

  def _update_11ax_configs(self, wifi_config: wifi_configs.WiFiConfig):
    """Updates 802.11AX related configurations."""
    if wifi_config.standard not in constants.STANDARDS_SUPPORT_HE_CAPAB:
      return
    self.update('ieee80211ax', '1')

    width = typing.cast(wifi_configs.ChannelWidth, wifi_config.width)
    self.update('he_oper_chwidth', width.to_hostapd_enum())

    if wifi_config.band_type == wifi_configs.BandType.BAND_2G:
      return

    if wifi_config.width == wifi_configs.ChannelWidth.WIDTH_80:
      self.update(
          'he_oper_centr_freq_seg0_idx',
          str(_get_center_channel_with_width_80mhz(wifi_config.channel)),
      )

  def _update_encryption_configs(
      self, wifi_config: wifi_configs.WiFiConfig
  ):
    """Updates the encryption configuration."""
    # Feed this to the encryption config checking.
    if wifi_config.pmf is not None:
      self.update('ieee80211w', str(wifi_config.pmf.value))
    wifi_config.encryption_config.update_hostapd_conf(hostapd_conf=self)


class HostapdManager:
  """The class for managing one hostapd instance on the AP device."""

  def __init__(
      self,
      device: 'OpenWrtDevice',
      wifi_id: int,
      phy: iw_utils.Phy,
      interface: str,
      wifi_config: wifi_configs.WiFiConfig,
      base_logger: (
          logging.Logger | mobly_logger.PrefixLoggerAdapter | None
      ) = None,
  ):
    """Constructor.

    Args:
      device: The AP device controller object.
      wifi_id: The unique ID of the WiFi network to start.
      phy: The wireless hardware device that the WiFi network is using.
      interface: The name of the network interface that the WiFi network is
        using.
      wifi_config: The WiFi configurations.
      base_logger: The base logger. Based on that logger, this class will prefix
        each log entry with string "[HostapdManager]".
    """
    self._device = device
    self._wifi_id = wifi_id
    self._wifi_config = wifi_config
    self._phy = phy
    self._interface = interface

    self._hostapd_config = None
    self._wifi_info = None
    self._remote_process = None
    self._identifier = f'wifi{self._wifi_id},{self._interface}'

    base_logger = base_logger or device.log
    self._log = mobly_logger.PrefixLoggerAdapter(
        base_logger,
        {
            mobly_logger.PrefixLoggerAdapter.EXTRA_KEY_LOG_PREFIX: (
                f'[HostapdManager|wifi{self._wifi_id}]'
            )
        },
    )
    self._local_work_dir = self._device.log_path
    self._remote_work_dir = self._device.remote_work_dir

  def start(self) -> wifi_configs.WifiInfo:
    """Starts this hostapd manager instance."""
    self._log.debug('Starting a remote hostapd instance.')
    try:
      return self._start()
    except (errors.BaseError, ssh_lib.SSHRemoteError):
      self._log.error('Stopping due to start failure.')
      with contextlib.suppress(errors.BaseError, ssh_lib.SSHRemoteError):
        self.stop()
      raise

  def _start(self) -> wifi_configs.WifiInfo:
    """Starts a hostapd process using given configs on the AP device."""
    dfs_channels = iw_utils.get_all_dfs_channels(self._phy)
    self._log.debug(
        'All DFS channels on phy %s: %s', self._phy.name, sorted(dfs_channels)
    )

    self._hostapd_config = HostapdConfig(
        interface=self._interface,
        dfs_channels=dfs_channels,
        is_custom_openwrt=utils.is_using_custom_image(self._device),
    )
    self._hostapd_config.update_from_wifi_config(self._wifi_config)

    self._wifi_info = wifi_configs.WifiInfo(
        id=self._wifi_id,
        ssid=self._hostapd_config.ssid,
        password=self._hostapd_config.password,
        interface=self._interface,
        phy_name=self._phy.name,
    )

    conf_remote_path = self._generate_remote_config_file()
    self._start_hostpad_process(conf_remote_path)
    return self._wifi_info

  def _generate_remote_config_file(self) -> str:
    """Generates the hostapd config file on the AP device."""
    filename = self._get_conf_filename()
    local_path = self._get_local_path(filename)
    remote_path = self._get_remote_path(filename)

    self._hostapd_config = typing.cast(HostapdConfig, self._hostapd_config)
    config_content = self._hostapd_config.config_content
    self._hostapd_config.write_to_file(local_path, config_content)
    self._device.push_file(local_path, remote_path)

    # Rename the local file so it can be directly opened on Sponge.
    os.rename(local_path, f'{local_path}.txt')
    return remote_path

  def _start_hostpad_process(
      self,
      conf_remote_path: str,
  ):
    """Starts the hostapd process and waits until the WiFi is ready."""
    log_file_path = self._get_local_path(self._get_log_filename())
    command = constants.Commands.HOSTAPD_START.format(
        conf_path=conf_remote_path,
    )
    proc = self._device.ssh.start_remote_process(
        command, get_pty=True, output_file_path=log_file_path
    )
    self._remote_process = proc

    self._hostapd_config = typing.cast(HostapdConfig, self._hostapd_config)
    wait_timeout = (
        _WIFI_START_WAIT_TIME_DFS
        if self._hostapd_config.is_using_a_dfs_channel()
        else _WIFI_START_WAIT_TIME
    )
    if not utils.wait_for_predicate(
        predicate=self._is_wifi_ready,
        timeout=wait_timeout,
        interval=_WIFI_START_CHECK_INTERVAL,
    ):
      raise errors.HostapdStartError(
          'Failed to start hostapd. Please check the hostapd log'
          f' {self._get_log_filename()} and config {self._get_conf_filename()}.'
      )

    self._log.debug('Started remote hostapd process.')

  def _is_wifi_ready(self):
    """Returns whether the WiFi is ready."""
    if self._remote_process is None:
      raise errors.HostapdStartError(
          'Hostapd process is not set. Please check whether hostapd object'
          ' has not been started or is already stopped.'
      )
    if self._remote_process.poll() is not None:
      raise errors.HostapdStartError(
          'Hostapd process has exited unexpectedly on the AP device. Please'
          f' check the hostapd log file {self._get_log_filename()} and conf'
          f' file {self._get_conf_filename()}'
      )

    stdout = self._device.ssh.execute_command(
        command=constants.Commands.IP_LINK_SHOW.format(
            interface=self._interface
        ),
        timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
        ignore_error=True,
    )
    return 'state UP' in stdout

  def __del__(self):
    self.stop()

  def stop(self):
    """Stops the remote hostapd process."""
    if self._remote_process is None:
      return

    self._log.debug(
        'Stopping hostapd process %d.',
        self._remote_process.pid,
    )
    proc = self._remote_process
    self._remote_process = None
    proc.terminate(
        timeout=_WIFI_STOP_WAIT_TIME.total_seconds(), assert_process_exit=True
    )

  def _get_remote_path(self, filename: str) -> str:
    return os.path.join(self._remote_work_dir, filename)

  def _get_local_path(self, filename: str) -> str:
    return os.path.join(self._local_work_dir, filename)

  def _get_conf_filename(self) -> str:
    return f'{self._identifier},hostapd.conf'

  def _get_log_filename(self) -> str:
    return f'{self._identifier},hostapd.log'
